import argparse
from argparse import ArgumentParser
from schnetpack.datasets import (
    QM9,
    ANI1,
    MD17,
    OrganicMaterialsDatabase,
    MaterialsProject,
)


class StoreDictKeyPair(argparse.Action):
    """
    From https://stackoverflow.com/a/42355279
    """

    def __init__(self, option_strings, dest, nargs=None, val_type=str, **kwargs):
        self._nargs = nargs
        self.val_type = val_type
        super(StoreDictKeyPair, self).__init__(
            option_strings, dest, nargs=nargs, **kwargs
        )

    def __call__(self, parser, namespace, values, option_string=None):
        my_dict = {}
        for kv in values:
            k, v = kv.split("=")
            # typecast
            if self.val_type == int:
                v = int(float(v))
            else:
                v = self.val_type(v)
            my_dict[k] = v
        setattr(namespace, self.dest, my_dict)


def get_mode_parsers():
    # json parser
    json_parser = ArgumentParser(add_help=False)
    json_parser.add_argument(
        "json_path",
        type=str,
        help="Path to argument file. (default: %(default)s)",
        default=None,
    )

    # train parser
    train_parser = ArgumentParser(add_help=False)
    train_parser.add_argument("datapath", help="Path to dataset")
    train_parser.add_argument("modelpath", help="Path of stored model")
    train_parser.add_argument(
        "--cuda", help="Set flag to use GPU(s) for training", action="store_true"
    )
    train_parser.add_argument(
        "--parallel",
        help="Run data-parallel on all available GPUs (specify with environment"
        " variable CUDA_VISIBLE_DEVICES)",
        action="store_true",
    )
    train_parser.add_argument(
        "--seed", type=int, default=None, help="Set random seed for torch and numpy."
    )
    train_parser.add_argument(
        "--overwrite", help="Remove previous model directory.", action="store_true"
    )

    # data split
    train_parser.add_argument(
        "--split_path", help="Path / destination of npz with data splits", default=None
    )
    train_parser.add_argument(
        "--split",
        help="Split into [train] [validation] and use remaining for testing",
        type=int,
        nargs=2,
        default=[None, None],
    )
    train_parser.add_argument(
        "--max_epochs",
        type=int,
        help="Maximum number of training epochs (default: %(default)s)",
        default=5000,
    )
    train_parser.add_argument(
        "--max_steps",
        type=int,
        help="Maximum number of training steps (default: %(default)s)",
        default=None,
    )
    train_parser.add_argument(
        "--lr",
        type=float,
        help="Initial learning rate (default: %(default)s)",
        default=1e-4,
    )
    train_parser.add_argument(
        "--lr_patience",
        type=int,
        help="Epochs without improvement before reducing the learning rate "
        "(default: %(default)s)",
        default=25,
    )
    train_parser.add_argument(
        "--lr_decay",
        type=float,
        help="Learning rate decay (default: %(default)s)",
        default=0.8,
    )
    train_parser.add_argument(
        "--lr_min",
        type=float,
        help="Minimal learning rate (default: %(default)s)",
        default=1e-6,
    )

    train_parser.add_argument(
        "--logger",
        help="Choose logger for training process (default: %(default)s)",
        choices=["csv", "tensorboard"],
        default="csv",
    )
    train_parser.add_argument(
        "--log_every_n_epochs",
        type=int,
        help="Log metrics every given number of epochs (default: %(default)s)",
        default=1,
    )
    train_parser.add_argument(
        "--n_epochs",
        type=int,
        help="Maximum number of training epochs (default: %(default)s)",
        default=1000,
    )
    train_parser.add_argument(
        "--checkpoint_interval",
        type=int,
        help="Store checkpoint every n epochs (default: %(default)s)",
        default=1,
    )
    train_parser.add_argument(
        "--keep_n_checkpoints",
        type=int,
        help="Number of checkpoints that will be stored (default: %(default)s)",
        default=3,
    )
    train_parser.add_argument(
        "--environment_provider_device",
        help="Choose device for environment providers. It is recommended to keep CPU. (default: %(default)s)",
        choices=["cpu", "cuda"],
        default="cpu",
    )

    # evaluation parser
    eval_parser = ArgumentParser(add_help=False)
    eval_parser.add_argument("datapath", help="Path to dataset")
    eval_parser.add_argument("modelpath", help="Path of stored model")
    eval_parser.add_argument(
        "--cuda", help="Set flag to use GPU(s) for evaluation", action="store_true"
    )
    eval_parser.add_argument(
        "--parallel",
        help="Run data-parallel on all available GPUs (specify with environment"
        " variable CUDA_VISIBLE_DEVICES)",
        action="store_true",
    )
    eval_parser.add_argument(
        "--batch_size",
        type=int,
        help="Mini-batch size for evaluation (default: %(default)s)",
        default=100,
    )
    eval_parser.add_argument(
        "--split",
        help="Evaluate trained model on given split",
        choices=["train", "validation", "test"],
        default=["test"],
        nargs="+",
    )
    eval_parser.add_argument(
        "--overwrite", help="Remove previous evaluation files", action="store_true"
    )
    eval_parser.add_argument(
        "--environment_provider_device",
        help="Choose device for environment providers. It is recommended to keep CPU. (default: %(default)s)",
        choices=["cpu", "cuda"],
        default="cpu",
    )

    return json_parser, train_parser, eval_parser


def get_model_parsers():
    # model parsers
    schnet_parser = ArgumentParser(add_help=False)
    schnet_parser.add_argument(
        "--features", type=int, help="Size of atom-wise representation", default=128
    )
    schnet_parser.add_argument(
        "--interactions", type=int, help="Number of interaction blocks", default=6
    )
    schnet_parser.add_argument(
        "--cutoff_function",
        help="Functional form of the cutoff",
        choices=["hard", "cosine", "mollifier"],
        default="cosine",
    )
    schnet_parser.add_argument(
        "--num_gaussians",
        type=int,
        default=50,
        help="Number of Gaussians to expand distances (default: %(default)s)",
    )
    schnet_parser.add_argument(
        "--normalize_filter",
        action="store_true",
        help="Normalize convolution filters by number of neighbors",
    )

    wacsf_parser = ArgumentParser(add_help=False)
    wacsf_parser.add_argument(
        "--radial",
        type=int,
        default=22,
        help="Number of radial symmetry functions (default: %(default)s)",
    )
    wacsf_parser.add_argument(
        "--angular",
        type=int,
        default=5,
        help="Number of angular symmetry functions (default: %(default)s)",
    )
    wacsf_parser.add_argument(
        "--zetas",
        type=int,
        nargs="+",
        default=[1],
        help="List of zeta exponents used for angle resolution (default: %(default)s)",
    )
    wacsf_parser.add_argument(
        "--standardize",
        action="store_true",
        help="Standardize wACSF before atomistic network.",
    )
    # Atomistic network parameters
    wacsf_parser.add_argument(
        "--n_nodes",
        type=int,
        default=100,
        help="Number of nodes in atomic networks (default: %(default)s)",
    )
    wacsf_parser.add_argument(
        "--n_layers",
        type=int,
        default=2,
        help="Number of layers in atomic networks (default: %(default)s)",
    )
    # Advances wACSF settings
    wacsf_parser.add_argument(
        "--centered",
        action="store_true",
        help="Use centered Gaussians for radial functions",
    )
    wacsf_parser.add_argument(
        "--crossterms", action="store_true", help="Use crossterms in angular functions"
    )
    wacsf_parser.add_argument(
        "--behler", action="store_true", help="Switch to conventional ACSF"
    )
    wacsf_parser.add_argument(
        "--elements",
        default=["H", "C", "N", "O", "F"],
        nargs="+",
        help="List of elements to be used for symmetry functions "
        "(default: %(default)s).",
    )

    return schnet_parser, wacsf_parser


def get_data_parsers():
    # qm9
    qm9_parser = ArgumentParser(add_help=False)
    qm9_parser.add_argument(
        "--property",
        type=str,
        help="Database property to be predicted (default: %(default)s)",
        default=QM9.U0,
        choices=[
            QM9.A,
            QM9.B,
            QM9.C,
            QM9.mu,
            QM9.alpha,
            QM9.homo,
            QM9.lumo,
            QM9.gap,
            QM9.r2,
            QM9.zpve,
            QM9.U0,
            QM9.U,
            QM9.H,
            QM9.G,
            QM9.Cv,
        ],
    )
    qm9_parser.add_argument(
        "--cutoff",
        type=float,
        default=10.0,
        help="Cutoff radius of local environment (default: %(default)s)",
    )
    qm9_parser.add_argument(
        "--batch_size",
        type=int,
        help="Mini-batch size for training (default: %(default)s)",
        default=100,
    )
    qm9_parser.add_argument(
        "--environment_provider",
        type=str,
        default="simple",
        choices=["simple", "ase", "torch"],
        help="Environment provider for dataset. (default: %(default)s)",
    )
    qm9_parser.add_argument(
        "--remove_uncharacterized",
        help="Remove uncharacterized molecules from QM9 (default: %(default)s)",
        action="store_true",
    )

    ani1_parser = ArgumentParser(add_help=False)
    ani1_parser.add_argument(
        "--property",
        type=str,
        help="Database property to be predicted (default: %(default)s)",
        default=ANI1.energy,
        choices=[ANI1.energy],
    )
    ani1_parser.add_argument(
        "--cutoff",
        type=float,
        default=10.0,
        help="Cutoff radius of local environment (default: %(default)s)",
    )
    ani1_parser.add_argument(
        "--batch_size",
        type=int,
        help="Mini-batch size for training (default: %(default)s)",
        default=100,
    )
    ani1_parser.add_argument(
        "--environment_provider",
        type=str,
        default="simple",
        choices=["simple", "ase", "torch"],
        help="Environment provider for dataset. (default: %(default)s)",
    )
    ani1_parser.add_argument(
        "--num_heavy_atoms",
        type=int,
        help="Number of heavy atoms that will be loaded into the database."
        " (default: %(default)s)",
        default=8,
    )
    matproj_parser = ArgumentParser(add_help=False)
    matproj_parser.add_argument(
        "--property",
        type=str,
        help="Database property to be predicted" " (default: %(default)s)",
        default=MaterialsProject.EformationPerAtom,
        choices=[
            MaterialsProject.EformationPerAtom,
            MaterialsProject.EPerAtom,
            MaterialsProject.BandGap,
            MaterialsProject.TotalMagnetization,
        ],
    )
    matproj_parser.add_argument(
        "--cutoff",
        type=float,
        default=5.0,
        help="Cutoff radius of local environment (default: %(default)s)",
    )
    matproj_parser.add_argument(
        "--batch_size",
        type=int,
        help="Mini-batch size for training (default: %(default)s)",
        default=32,
    )
    matproj_parser.add_argument(
        "--environment_provider",
        type=str,
        default="torch",
        choices=["simple", "ase", "torch"],
        help="Environment provider for dataset. (default: %(default)s)",
    )
    matproj_parser.add_argument(
        "--apikey",
        help="API key for Materials Project (see https://materialsproject.org/open)",
        default=None,
    )
    matproj_parser.add_argument(
        "--timestamp",
        help="Timestamp at which to reconstruct the dataset",
        default="2017-12-04 14:20",
    )

    md17_parser = ArgumentParser(add_help=False)
    md17_parser.add_argument(
        "--property",
        type=str,
        help="Database property to be predicted" " (default: %(default)s)",
        default=MD17.energy,
        choices=[MD17.energy],
    )
    md17_parser.add_argument(
        "--cutoff",
        type=float,
        default=5.0,
        help="Cutoff radius of local environment (default: %(default)s)",
    )
    md17_parser.add_argument(
        "--batch_size",
        type=int,
        help="Mini-batch size for training (default: %(default)s)",
        default=100,
    )
    md17_parser.add_argument(
        "--environment_provider",
        type=str,
        default="simple",
        choices=["simple", "ase", "torch"],
        help="Environment provider for dataset. (default: %(default)s)",
    )
    md17_parser.add_argument(
        "--ignore_forces", action="store_true", help="Ignore forces during training."
    )
    md17_parser.add_argument(
        "--molecule",
        type=str,
        help="Choose molecule inside the MD17 dataset. (default: %(default)s)",
        default="ethanol",
        choices=MD17.datasets_dict.keys(),
    )
    md17_parser.add_argument(
        "--rho",
        type=float,
        help="Energy-force trade-off. For rho=0, use forces only. "
        "(default: %(default)s)",
        default=0.1,
    )
    omdb_parser = ArgumentParser(add_help=False)
    omdb_parser.add_argument(
        "--property",
        type=str,
        help="Database property to be predicted (default: %(default)s)",
        default=OrganicMaterialsDatabase.BandGap,
        choices=[OrganicMaterialsDatabase.BandGap],
    )
    omdb_parser.add_argument(
        "--cutoff",
        type=float,
        default=5.0,
        help="Cutoff radius of local environment (default: %(default)s)",
    )
    omdb_parser.add_argument(
        "--batch_size",
        type=int,
        help="Mini-batch size for training (default: %(default)s)",
        default=32,
    )
    omdb_parser.add_argument(
        "--environment_provider",
        type=str,
        default="torch",
        choices=["simple", "ase", "torch"],
        help="Environment provider for dataset. (default: %(default)s)",
    )
    custom_data_parser = ArgumentParser(add_help=False)
    custom_data_parser.add_argument(
        "--property",
        type=str,
        help="Database property to be predicted (default: %(default)s)",
        default="energy",
    )
    custom_data_parser.add_argument(
        "--cutoff",
        type=float,
        default=10.0,
        help="Cutoff radius of local environment (default: %(default)s)",
    )
    custom_data_parser.add_argument(
        "--batch_size",
        type=int,
        help="Mini-batch size for training (default: %(default)s)",
        default=100,
    )
    custom_data_parser.add_argument(
        "--environment_provider",
        type=str,
        default="simple",
        choices=["simple", "ase", "torch"],
        help="Environment provider for dataset. (default: %(default)s)",
    )
    custom_data_parser.add_argument(
        "--derivative",
        type=str,
        help="Derivative of dataset property to be predicted (default: %(default)s)",
        default=None,
    )
    custom_data_parser.add_argument(
        "--negative_dr",
        action="store_true",
        help="Multiply derivatives with -1 for training. (default: %(default)s)",
    )
    custom_data_parser.add_argument(
        "--force",
        type=str,
        help="Name of force property in database. Alias for‚ derivative + setting "
        "negative_dr. (default: %(default)s)",
        default=None,
    )
    custom_data_parser.add_argument(
        "--contributions",
        type=str,
        help="Contributions of dataset property to be predicted (default: %(default)s)",
        default=None,
    )
    custom_data_parser.add_argument(
        "--stress",
        type=str,
        help="Train on stress tensor if not None (default: %(default)s)",
        default=None,
    )
    custom_data_parser.add_argument(
        "--aggregation_mode",
        type=str,
        help="Select mode for aggregating atomic properties. (default: %(default)s)",
        default="sum",
    )
    custom_data_parser.add_argument(
        "--output_module",
        type=str,
        help="Select matching output module for selected property. (default: %("
        "default)s)",
        default="atomwise",
        choices=[
            "atomwise",
            "elemental_atomwise",
            "dipole_moment",
            "elemental_dipole_moment",
            "polarizability",
            "isotropic_polarizability",
            "electronic_spatial_extent",
        ],
    )
    custom_data_parser.add_argument(
        "--rho",
        action=StoreDictKeyPair,
        nargs="+",
        metavar="KEY=VAL",
        help="Define loss tradeoff weights with prop=weight. (default: %(default)s)",
        default=dict(),
        val_type=float,
    )
    return (
        qm9_parser,
        ani1_parser,
        matproj_parser,
        md17_parser,
        omdb_parser,
        custom_data_parser,
    )


def build_parser():
    main_parser = ArgumentParser()
    # get parsers
    json_parser, train_parser, eval_parser = get_mode_parsers()
    schnet_parser, wacsf_parser = get_model_parsers()
    (
        qm9_parser,
        ani1_parser,
        matproj_parser,
        md17_parser,
        omdb_parser,
        custom_data_parser,
    ) = get_data_parsers()

    # subparser structure
    # mode
    mode_subparsers = main_parser.add_subparsers(dest="mode", help="main arguments")
    mode_subparsers.required = True
    train_subparser = mode_subparsers.add_parser("train", help="training help")
    eval_subparser = mode_subparsers.add_parser(
        "eval", help="evaluation help", parents=[eval_parser]
    )
    json_subparser = mode_subparsers.add_parser(
        "from_json", help="load from json help", parents=[json_parser]
    )

    # train mode
    train_subparsers = train_subparser.add_subparsers(
        dest="model", help="Model-specific arguments"
    )
    train_subparsers.required = True
    # model
    schnet_subparser = train_subparsers.add_parser("schnet", help="SchNet help")
    wacsf_subparser = train_subparsers.add_parser("wacsf", help="wacsf help")

    # schnet
    schnet_subparsers = schnet_subparser.add_subparsers(
        dest="dataset", help="Dataset specific arguments"
    )
    schnet_subparsers.required = True
    schnet_subparsers.add_parser(
        "ani1",
        help="ANI1 dataset help",
        parents=[train_parser, schnet_parser, ani1_parser],
    )
    schnet_subparsers.add_parser(
        "matproj",
        help="Materials Project dataset help",
        parents=[train_parser, schnet_parser, matproj_parser],
    )

    schnet_subparsers.add_parser(
        "md17",
        help="MD17 dataset help",
        parents=[train_parser, schnet_parser, md17_parser],
    )
    schnet_subparsers.add_parser(
        "omdb",
        help="Organic Materials dataset help",
        parents=[train_parser, schnet_parser, omdb_parser],
    )

    schnet_subparsers.add_parser(
        "qm9",
        help="QM9 dataset help",
        parents=[train_parser, schnet_parser, qm9_parser],
    )
    schnet_subparsers.add_parser(
        "custom",
        help="Custom dataset help",
        parents=[train_parser, schnet_parser, custom_data_parser],
    )

    # wacsf
    wacsf_subparsers = wacsf_subparser.add_subparsers(
        dest="dataset", help="Dataset specific arguments"
    )
    wacsf_subparsers.required = True
    wacsf_subparsers.add_parser(
        "ani1",
        help="ANI1 dataset help",
        parents=[train_parser, wacsf_parser, ani1_parser],
    )
    wacsf_subparsers.add_parser(
        "matproj",
        help="Materials Project dataset help",
        parents=[train_parser, wacsf_parser, matproj_parser],
    )
    wacsf_subparsers.add_parser(
        "md17",
        help="MD17 dataset help",
        parents=[train_parser, wacsf_parser, md17_parser],
    )
    wacsf_subparsers.add_parser(
        "omdb",
        help="Organic Materials dataset help",
        parents=[train_parser, wacsf_parser, omdb_parser],
    )
    wacsf_subparsers.add_parser(
        "qm9", help="QM9 dataset help", parents=[train_parser, wacsf_parser, qm9_parser]
    )
    wacsf_subparsers.add_parser(
        "custom",
        help="Custom dataset help",
        parents=[train_parser, wacsf_parser, custom_data_parser],
    )
    return main_parser


if __name__ == "__main__":
    parser = build_parser()
    args = parser.parse_args()
    print(args)



"""
This module contains various utility functions and classes used in combination with :obj:`schnetpack.md`.
This includes e.g. unit conversion, loading of molecular dynamics data and the computation of spectra.
"""
from .basic_utils import *
from .md_units import *
from .hdf5_data import *
from .normal_mode_transformation import *
from .spectra import *
from .thermostat_utils import *



"""
Integrators are used to propagate the simulated system in time. SchNetPack
provides two basic types of integrators. The Velocity Verlet integrator is a standard
integrator for a purely classical simulations of the nuclei. The ring polymer molecular dynamics
integrator simulates multiple replicas of the system coupled by harmonic springs and recovers
a certain extent of nuclear quantum effects (e.g. tunneling).
"""
import torch
import numpy as np

from schnetpack.md.utils import NormalModeTransformer, MDUnits

__all__ = ["VelocityVerlet", "RingPolymer", "NPTVelocityVerlet", "NPTRingPolymer"]


class Integrator:
    """
    Basic integrator class template. Uses the typical scheme of propagating
    system momenta in two half steps and system positions in one main step.
    The half steps are defined by default and only the _main_step function
    needs to be specified. Uses atomic time units internally.

    If required, the torch graphs generated by this routine can be detached
    every step via the detach flag.

    Args:
        time_step (float): Integration time step in femto seconds.
        detach (bool): If set to true, torch graphs of the propagation are
                       detached after every step (recommended, due to extreme
                       memory usage). This functionality could in theory be used
                       to formulate differentiable MD.
    """

    def __init__(self, time_step, detach=True, device="cuda"):
        self.time_step = time_step * MDUnits.fs2internal
        self.detach = detach
        self.device = device

    def main_step(self, system):
        """
        Main integration step wrapper routine to make a default detach
        behavior possible. Calls upon _main_step to perform the actual
        propagation of the system.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        self._main_step(system)
        if self.detach:
            system.positions = system.positions.detach()
            system.momenta = system.momenta.detach()

    def half_step(self, system):
        """
        Half steps propagating the system momenta according to:

        ..math::
            p = p + \frac{1}{2} F \delta t

        Args:
            system (object): System class containing all molecules and their
                             replicas.
        """
        system.momenta = system.momenta + 0.5 * system.forces * self.time_step

        if self.detach:
            system.momenta = system.momenta.detach()

    def _main_step(self, system):
        """
        Main integration step to be implemented in derived routines.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        raise NotImplementedError


class VelocityVerlet(Integrator):
    """
    Standard velocity Verlet integrator for non ring-polymer simulations.

    Args:
        time_step (float): Integration time step in femto seconds.
    """

    def __init__(self, time_step, device="cuda"):
        super(VelocityVerlet, self).__init__(time_step, device=device)

    def _main_step(self, system):
        """
        Propagate the positions of the system according to:

        ..math::
            q = q + \frac{p}{m} \delta t

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        system.positions = (
            system.positions + self.time_step * system.momenta / system.masses
        )

        if self.detach:
            system.positions = system.positions.detach()


class RingPolymer(Integrator):
    """
    Integrator for ring polymer molecular dynamics, as e.g. described in
    [#rpmd1]_

    During the main step, ring polymer positions and momenta are transformed
    from bead to normal mode representation, propagated deterministically and
    then transformed back. Needs the number of beads and the ring polymer
    temperature in order to initialize the propagator matrix. The integrator
    reverts to standard velocity Verlet integration if only one bead is used.

    Uses atomic units of time internally.

    Args:
        n_beads (int): Number of beads in the ring polymer.
        time_step (float): Time step in femto seconds.
        temperature (float): Ring polymer temperature in Kelvin.
        transformation (schnetpack.md.utils.NormalModeTransformer): Normal mode transformer class.
        device (str): Device used for computations, default is GPU ('cuda')

    References
    ----------
    .. [#rpmd1] Ceriotti, Parrinello, Markland, Manolopoulos:
       Efficient stochastic thermostatting of path integral molecular dynamics.
       The Journal of Chemical Physics, 133, 124105. 2010.
    """

    def __init__(
        self,
        n_beads,
        time_step,
        temperature,
        transformation=NormalModeTransformer,
        device="cuda",
    ):
        super(RingPolymer, self).__init__(time_step, device=device)

        self.n_beads = n_beads

        # Compute the ring polymer frequency
        self.omega = MDUnits.kB * n_beads * temperature / MDUnits.hbar
        self.transformation = transformation(n_beads, device=self.device)

        # Set up omega_normal, the ring polymer frequencies in normal mode
        # representation
        self.omega_normal = (
            2
            * self.omega
            * torch.sin(
                torch.arange(self.n_beads, device=device).float() * np.pi / self.n_beads
            )
        )

        # Initialize the propagator matrices
        self.propagator = self._init_propagator()

    def _init_propagator(self):
        """
        Constructs the ring polymer propagator in normal mode representation
        as for example given in [#rpmd2]_.

        Returns:
            torch.Tensor: Propagator with the dimension n_beads x 2 x 2,
                          where the last two dimensions mix the systems
                          momenta and positions in normal mode representation.

        References
        ----------
        .. [#rpmd2] Ceriotti, Parrinello, Markland, Manolopoulos:
           Efficient stochastic thermostatting of path integral molecular
           dynamics.
           The Journal of Chemical Physics, 133, 124105. 2010.
        """

        # Compute basic terms
        omega_dt = self.omega_normal * self.time_step
        cos_dt = torch.cos(omega_dt)
        sin_dt = torch.sin(omega_dt)

        # Initialize the propagator
        propagator = torch.zeros(self.n_beads, 2, 2, device=self.device)

        # Define the propagator elements, the central normal mode is treated
        # special
        propagator[:, 0, 0] = cos_dt
        propagator[:, 1, 1] = cos_dt
        propagator[:, 0, 1] = -sin_dt * self.omega_normal
        propagator[1:, 1, 0] = sin_dt[1:] / self.omega_normal[1:]

        # Centroid normal mode is special as reverts to standard velocity
        # Verlet for one bead.
        propagator[0, 1, 0] = self.time_step

        # Expand dimensions to avoid broadcasting in main_step
        propagator = propagator[..., None, None, None]

        return propagator

    def _main_step(self, system):
        """
        Main propagation step for ring polymer dynamics. First transforms
        positions and momenta to their normal mode representations,
        then applies the propagator defined above (mixing momenta and
        positions accordingly) and performs a backtransformation to the bead
        momenta and positions afterwards, which are used to update the
        current system state.

        Args:
            system (object): System class containing all molecules and their
                             replicas.
        """
        # Transform to normal mode representation
        positions_normal = self.transformation.beads2normal(system.positions)
        momenta_normal = self.transformation.beads2normal(system.momenta)

        # Propagate ring polymer
        momenta_normal = (
            self.propagator[:, 0, 0] * momenta_normal
            + self.propagator[:, 0, 1] * positions_normal * system.masses
        )
        positions_normal = (
            self.propagator[:, 1, 0] * momenta_normal / system.masses
            + self.propagator[:, 1, 1] * positions_normal
        )

        # Transform back to bead representation
        system.positions = self.transformation.normal2beads(positions_normal)
        system.momenta = self.transformation.normal2beads(momenta_normal)

        if self.detach:
            system.positions = system.positions.detach()
            system.momenta = system.momenta.detach()


class NPTVelocityVerlet(VelocityVerlet):
    """
    Verlet integrator for constant pressure dynamics (NPT). Since barostats modify the position update,
    a routine defined in the respectve barostat class is called every main step.

    Args:
        time_step (float): Integration time step in femto seconds.
        barostat (schnetpack.md.simulation_hooks.BarostatHook): Barostat used for constant pressure dynamics.
    """

    def __init__(self, time_step, barostat, device="cuda"):
        super(NPTVelocityVerlet, self).__init__(time_step, device=device)
        self.barostat = barostat

    def _main_step(self, system):
        """
        Main integrator step, where the barostat routine is used to propagate the system positions and cells.
        """
        self.barostat.propagate_system(system)

        if self.detach:
            system.positions = system.positions.detach()
            system.cells = system.cells.detach()


class NPTRingPolymer(RingPolymer):
    """
    Ring polymer integrator for constant pressure dynamics (NPT). Here, the barostat modifies the main and the
    half steps.

    Args:
        n_beads (int): Number of beads in the ring polymer.
        time_step (float): Time step in femto seconds.
        temperature (float): Ring polymer temperature in Kelvin.
        barostat (schnetpack.md.simulation_hooks.BarostatHook): Barostat used for constant pressure dynamics.
        transformation (schnetpack.md.utils.NormalModeTransformer): Normal mode transformer class.
        device (str): Device used for computations, default is GPU ('cuda')
    """

    def __init__(
        self,
        n_beads,
        time_step,
        temperature,
        barostat,
        transformation=NormalModeTransformer,
        device="cuda",
    ):
        super(NPTRingPolymer, self).__init__(
            n_beads,
            time_step,
            temperature,
            transformation=transformation,
            device=device,
        )
        self.barostat = barostat

    def half_step(self, system):
        """
        Half steps propagating the system and barostat momenta.

        Args:
            system (object): System class containing all molecules and their
                             replicas.
        """
        self.barostat.propagate_barostat_half_step(system)

        system.momenta = system.momenta + 0.5 * system.forces * self.time_step

        if self.detach:
            system.momenta = system.momenta.detach()

    def _main_step(self, system):
        """
        Perform the main update using the barostat routine.

        Args:
            system (object): System class containing all molecules and their
                             replicas.
        """
        self.barostat.propagate_system(system)

        if self.detach:
            system.positions = system.positions.detach()
            system.momenta = system.momenta.detach()
            system.cells = system.cells.detach()



import os
from ase.db import connect
from ase.io.extxyz import read_xyz
from tqdm import tqdm
import tempfile


def xyz_to_extxyz(
    xyz_path,
    extxyz_path,
    atomic_properties="Properties=species:S:1:pos:R:3",
    molecular_properties=[],
):
    """
    Convert a xyz-file to extxyz.

    Args:
        xyz_path (str): path to the xyz file
        extxyz_path (str): path to extxyz-file
        atomic_properties (str): property-string
        molecular_properties (list): molecular properties contained in the
            comment line
    """
    # ensure valid property string
    atomic_properties = parse_property_string(atomic_properties)
    new_file = open(extxyz_path, "w")
    with open(xyz_path, "r") as xyz_file:
        while True:
            first_line = xyz_file.readline()
            if first_line == "":
                break
            n_atoms = int(first_line.strip("\n"))
            molecular_data = xyz_file.readline().strip("/n").split()
            assert len(molecular_data) == len(molecular_properties), (
                "The number of datapoints and " "properties do not match!"
            )
            comment = " ".join(
                [
                    "{}={}".format(prop, val)
                    for prop, val in zip(molecular_properties, molecular_data)
                ]
            )
            new_file.writelines(str(n_atoms) + "\n")
            new_file.writelines(" ".join([atomic_properties, comment]) + "\n")
            for i in range(n_atoms):
                line = xyz_file.readline()
                new_file.writelines(line)
    new_file.close()


def extxyz_to_db(extxyz_path, db_path):
    r"""
    Convertes en extxyz-file to an ase database

    Args:
        extxyz_path (str): path to extxyz-file
        db_path(str): path to sqlite database
    """
    with connect(db_path, use_lock_file=False) as conn:
        with open(extxyz_path) as f:
            for at in tqdm(read_xyz(f, index=slice(None)), "creating ase db"):
                data = {}
                if at.has("forces"):
                    data["forces"] = at.get_forces()
                data.update(at.info)
                conn.write(at, data=data)


def xyz_to_db(
    xyz_path,
    db_path,
    atomic_properties="Properties=species:S:1:pos:R:3",
    molecular_properties=[],
):
    """
    Convertes a xyz-file to an ase database.

    Args:
        xyz_path (str): path to the xyz file
        db_path(str): path to sqlite database
        atomic_properties (str): property-string
        molecular_properties (list): molecular properties contained in the
            comment line
    """
    # build temp file in extended xyz format
    extxyz_path = os.path.join(tempfile.mkdtemp(), "temp.extxyz")
    xyz_to_extxyz(xyz_path, extxyz_path, atomic_properties, molecular_properties)
    # build database from extended xyz
    extxyz_to_db(extxyz_path, db_path)


def generate_db(
    file_path,
    db_path,
    atomic_properties="Properties=species:S:1:pos:R:3",
    molecular_properties=[],
):
    """
    Convert a file with molecular information to an ase database. Currently
    supports .xyz and .extxyz.

    Args:
        file_path (str): path to the input file
        db_path(str): path to sqlite database
        atomic_properties (str): property-string
        molecular_properties (list): molecular properties contained in the
            comment line
    """
    # check if file extension is valid
    filename, file_extension = os.path.splitext(file_path)
    if file_extension not in [".xyz", ".extxyz"]:
        raise NotImplementedError(
            "{} is not a supported file " "extension!".format(file_extension)
        )
    # check if file has property string
    with open(file_path, "r") as file:
        _ = file.readline()
        comment_line = file.readline()
    # build database file
    if "Properties=" in comment_line:
        extxyz_to_db(file_path, db_path)
    else:
        xyz_to_db(file_path, db_path, atomic_properties, molecular_properties)


def parse_property_string(prop_str):
    """
    Generate valid property string for extended xyz files.
    (ref. https://libatoms.github.io/QUIP/io.html#extendedxyz)

    Args:
        prop_str (str): Valid property string, or appendix of property string

    Returns:
        valid property string
    """
    if prop_str.startswith("Properties="):
        return prop_str
    return "Properties=species:S:1:pos:R:3:" + prop_str



"""
Calculators for molecular dynamics simulations in SchNetPack. These calculators take the current structures
in the :obj:`schnetpack.md.System` class and compute the associated forces. Other properties can be returned,
but molecular forces are the bare minimum for driving a simulations. All calculators should be derived from
the base classes :obj:`MDCalculator` (if no external code is called) or :obj:`QMCalculator`
(if external codes are required).

Currently implemented machine learning calculators include the :obj:`SchnetPackCalculator` for all models
generated with SchNetPack and the :obj:`SGDMLCalculator` for sGDML models [#sgdml5]_ .
In addition, an :obj:`OrcaCalculator` class can be used to carry out molecular dynamics using the ORCA
electronic structure code [#orca2]_ .

References
----------
.. [#sgdml5] Chmiela, Sauceda, Müller, Tkatchenko:
   Towards Exact Molecular Dynamics Simulations with Machine-Learned Force Fields.
   Nature Communications, 9 (1), 3887. 2018.
.. [#orca2] Neese:
   The ORCA program system.
   WIREs Comput Mol Sci, 2 (1), 73-78. 2012.
"""
from .basic_calculators import *
from .orca_calculator import *
from .schnet_calculator import *
from .sgdml_calculator import *
from .ensemble_calculator import *



"""
This module contains various thermostats for regulating the temperature of the system during
molecular dynamics simulations. Apart from standard thermostats for convetional simulations,
a series of special thermostats developed for ring polymer molecular dynamics is also provided.
"""
import torch

from schnetpack.md.utils import MDUnits, YSWeights, StableSinhDiv
from schnetpack.md.simulation_hooks import SimulationHook

__all__ = [
    "BarostatHook",
    "NHCBarostatIsotropic",
    "NHCBarostatAnisotropic",
    "RPMDBarostat",
]


class BarostatError(Exception):
    """
    Exception for barostat hooks.
    """

    pass


class BarostatHook(SimulationHook):
    """
    Basic barostat hook for simulator class. This class is initialized based on the simulator and system
    specifications during the first MD step. Barostats are applied before and after each MD step. In addition,
    they modify the update of positions and cells, which is why they have to be used with modified integrators.

    Args:
        target_pressure (float): Target pressure of the system (in bar).
        temperature_bath (float): Target temperature applied to the cell fluctuations.
        detach (bool): Whether the computational graph should be detached after each simulation step. Default is true,
                       should be changed if differentiable MD is desired.
    """

    def __init__(self, target_pressure, temperature_bath, detach=True):
        self.target_pressure = target_pressure * MDUnits.bar2internal
        self.temperature_bath = temperature_bath
        self.initialized = False
        self.device = None
        self.n_replicas = None
        self.n_molecules = None
        self.n_atoms = None
        self.detach = detach
        self.time_step = None

    def on_simulation_start(self, simulator):
        """
        Routine to initialize the barostat based on the current state of the simulator. Reads the device to be uses,
        as well as the number of molecular replicas present in simulator.system. A flag is set so that the barostat
        is not reinitialized upon continuation of the MD.

        Main function is the _init_barostat routine, which takes the simulator as input and must be provided for every
        new barostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                                the time step, system, etc.
        """
        self.device = simulator.system.device
        self.n_replicas = simulator.system.n_replicas
        self.n_molecules = simulator.system.n_molecules
        self.n_atoms = simulator.system.max_n_atoms
        self.time_step = simulator.integrator.time_step

        if not self.initialized:
            self._init_barostat(simulator)
            self.initialized = True

    def on_step_begin(self, simulator):
        """
        First application of the barostat before the first half step of the dynamics. Applies a mask to the system
        momenta in order to avoid problems of e.g. noise added to the zero padded tensors. The detach is carried out
        here.

        Main function is the _apply_barostat routine, which takes the simulator as input and must be provided for
        every new barostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                the time step, system, etc.
        """
        # Apply thermostat
        self._apply_barostat(simulator)

        # Re-apply atom masks for differently sized molecules, as some
        # thermostats add random noise
        simulator.system.momenta = (
            simulator.system.momenta * simulator.system.atom_masks
        )

        # Detach if requested
        if self.detach:
            simulator.system.momenta = simulator.system.momenta.detach()

    def on_step_end(self, simulator):
        """
        Second application of the barostat after the second half step of the dynamics. Applies a mask to the system
        momenta in order to avoid problems of e.g. noise added to the zero padded tensors. The detach is carried out
        here.

        Main function is the _apply_barostat routine, which takes the simulator as input and must be provided for
        every new barostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                the time step, system, etc.
        """
        # Apply thermostat
        self._apply_barostat(simulator)

        # Re-apply atom masks for differently sized molecules, as some
        # thermostats add random noise
        simulator.system.momenta = (
            simulator.system.momenta * simulator.system.atom_masks
        )

        # Detach if requested
        if self.detach:
            simulator.system.momenta = simulator.system.momenta.detach()

    def _init_barostat(self, simulator):
        """
        Dummy routine for initializing a barpstat based on the current simulator. Should be implemented for every
        new barostat. Has access to the information contained in the simulator class, e.g. number of replicas, time
        step, masses of the atoms, etc.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                the time step, system, etc.
        """
        raise NotImplementedError

    def _apply_barostat(self, simulator):
        """
        Dummy routine for applying the barostat to the system and propagating the barostat. Is called twice each
        simulation time step.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                the time step, system, etc.
        """
        raise NotImplementedError

    def propagate_system(self, system):
        """
        Propagate the system under the conditions imposed by the barostat. This has to be adapted to the specific
        barostat algorithm and should propagate the positons, cells and momenta (for RPMD). Defaults to classic Verlet.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        system.positions = (
            system.positions + self.time_step * system.momenta / system.masses
        )

    def propagate_barostat_half_step(self, system):
        """
        Routine only required for RPMD which propagates the barostat attached to the centroid cell. Is called during
        each half-step before system momenta are propagated.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        pass


class NHCBarostatIsotropic(BarostatHook):
    """
    Nose Hoover chain thermostat/barostat for isotropic cell fluctuations. This barostat already contains a built in
    thermostat, so no further temperature control is necessary. As suggested in [#nhc_barostat1]_, two separate chains
    are used to thermostat particle and cell momenta.

    Args:
        target_pressure (float): Target pressure in bar.
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Particle thermostat time constant in fs
        time_constant_cell (float): Cell thermostat time constant in fs. If None is given (default), the same time
                                    constant as for the thermostat component is used.
        time_constant_barostat (float): Barostat time constant in fs. If None is given (default), the same time constant
                                        as for the thermostat component is used.
        chain_length (int): Number of Nose-Hoover thermostats applied in the chain.
        multi_step (int): Number of steps used for integrating the NH equations of motion (default=2)
        integration_order (int): Order of the Yoshida-Suzuki integrator used for propagating the thermostat (default=3).
        massive (bool): Apply individual thermostat chains to all particle degrees of freedom (default=False).
        detach (bool): Whether the computational graph should be detached after each simulation step. Default is true,
                       should be changed if differentiable MD is desired.

    References
    ----------
    .. [#nhc_barostat1] Martyna, Tuckerman, Tobias, Klein:
       Explicit reversible integrators for extended systems dynamics.
       Molecular Physics, 87(5), 1117-1157. 1996.
    """

    def __init__(
        self,
        target_pressure,
        temperature_bath,
        time_constant,
        time_constant_cell=None,
        time_constant_barostat=None,
        chain_length=4,
        multi_step=4,
        integration_order=7,
        massive=False,
        detach=True,
    ):
        super(NHCBarostatIsotropic, self).__init__(
            target_pressure=target_pressure,
            temperature_bath=temperature_bath,
            detach=detach,
        )
        self.chain_length = chain_length
        self.frequency = 1 / (time_constant * MDUnits.fs2internal)

        if time_constant_cell is None:
            self.cell_frequency = self.frequency
        else:
            self.cell_frequency = 1 / (time_constant_cell * MDUnits.fs2internal)

        if time_constant_barostat is None:
            self.barostat_frequency = self.frequency
        else:
            self.barostat_frequency = 1 / (time_constant_barostat * MDUnits.fs2internal)

        # Compute kBT, since it will be used a lot
        self.kb_temperature = self.temperature_bath * MDUnits.kB

        # Stable sinh(x)/x approximation
        self.sinhdx = StableSinhDiv()

        # Propagation parameters
        self.multi_step = multi_step
        self.integration_order = integration_order
        self.ys_time_step = None

        # Find out number of particles
        self.degrees_of_freedom = None
        self.degrees_of_freedom_particles = None

        # Thermostat variables for particles
        self.t_velocities = None
        self.t_forces = None
        self.t_masses = None

        # Thermostat variables for cell
        self.t_velocities_cell = None
        self.t_forces_cell = None
        self.t_masses_cell = None

        # Barostat variables
        self.b_velocities_cell = None
        self.b_forces_cell = None
        self.b_masses_cell = None

        # Flag for massive theromstating
        self.massive = massive

    def _init_barostat(self, simulator):
        """
        Initialize the thermostat positions, forces, velocities and masses, as well as the number of degrees of freedom
        seen by each chain link. In the same manner, all quantities required for the barostat are initialized.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                    the time step, system, etc.
        """
        # Determine integration step via multi step and Yoshida Suzuki weights
        integration_weights = YSWeights(self.device).get_weights(self.integration_order)
        self.ys_time_step = (
            simulator.integrator.time_step * integration_weights / self.multi_step
        )

        # Determine internal degrees of freedom
        self.degrees_of_freedom = 3 * simulator.system.n_atoms.float()[None, :]

        # Determine degrees of freedom for particle thermostats (depends on massive)
        n_replicas, n_molecules, n_atoms, xyz = simulator.system.momenta.shape
        if self.massive:
            # Since momenta will be masked later, no need to set non-atoms to 0
            self.degrees_of_freedom_particles = torch.ones(
                (n_replicas, n_molecules, n_atoms, xyz), device=self.device
            )
        else:
            self.degrees_of_freedom_particles = self.degrees_of_freedom[None, None]

        # Set up internal variables
        self._init_thermostat_variables()
        self._init_barostat_variables()

    def _init_thermostat_variables(self):
        """
        Initialize all quantities required for the two thermostat chains on the particles and cells.
        """
        # Set up thermostat masses
        if self.massive:
            self.t_masses = torch.zeros(
                self.n_replicas,
                self.n_molecules,
                self.n_atoms,
                3,
                self.chain_length,
                device=self.device,
            )
        else:
            self.t_masses = torch.zeros(
                self.n_replicas,
                self.n_molecules,
                1,
                1,
                self.chain_length,
                device=self.device,
            )

        self.t_masses_cell = torch.zeros(
            self.n_replicas, self.n_molecules, self.chain_length, device=self.device
        )

        # Get masses of innermost thermostat
        self.t_masses[..., 0] = (
            self.degrees_of_freedom_particles
            * self.kb_temperature
            / self.frequency ** 2
        )
        self.t_masses_cell[..., 0] = (
            9.0 * self.kb_temperature / self.cell_frequency ** 2
        )
        # Set masses of remaining thermostats
        self.t_masses[..., 1:] = self.kb_temperature / self.frequency ** 2
        self.t_masses_cell[..., 1:] = self.kb_temperature / self.cell_frequency ** 2

        # Thermostat variables for particles
        self.t_velocities = torch.zeros_like(self.t_masses, device=self.device)
        self.t_forces = torch.zeros_like(self.t_masses, device=self.device)

        # Thermostat variables for cell
        self.t_velocities_cell = torch.zeros_like(
            self.t_masses_cell, device=self.device
        )
        self.t_forces_cell = torch.zeros_like(self.t_masses_cell, device=self.device)

        # Positions for conservation
        # self.t_positions = torch.zeros_like(self.t_masses, device=self.device)
        # self.t_positions_cell = torch.zeros_like(self.t_masses_cell, device=self.device)

    def _init_barostat_variables(self):
        """
        Initialize all quantities required for the barostat component.
        """
        # Set barostat masses
        self.b_masses_cell = torch.ones(
            self.n_replicas, self.n_molecules, device=self.device
        )
        self.b_masses_cell[:, :] = (
            (self.degrees_of_freedom + 3)
            * self.kb_temperature
            / self.barostat_frequency ** 2
        )
        # Remaining barostat variables
        self.b_velocities_cell = torch.zeros_like(
            self.b_masses_cell, device=self.device
        )
        self.b_forces_cell = torch.zeros_like(self.b_masses_cell, device=self.device)

    def _apply_barostat(self, simulator):
        """
        Apply the thermostat chains to the system and cell momenta. This is done by computing forces of the innermost
        thermostats, propagating the chain forward, updating the box velocities, particle momenta and associated
        energies. Based on this, the kinetic energies and forces can be updated, which are the propagated backward
        through the chain.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                                the time step, system, etc.
        """
        self._init_kinetic_energy(simulator.system)
        (
            kinetic_energy_for_thermostat,
            kinetic_energy_for_barostat,
        ) = self._compute_kinetic_energy(simulator.system)
        # kinetic_energy = self._compute_kinetic_energy(simulator.system)
        pressure, volume = self._compute_pressure(simulator.system)

        self._update_forces_barostat(kinetic_energy_for_barostat, pressure, volume)
        self._update_forces_thermostat(kinetic_energy_for_thermostat)

        for _ in range(self.multi_step):
            for idx_ys in range(self.integration_order):
                time_step = self.ys_time_step[idx_ys]

                self._chain_forward(time_step)
                self._update_box_velocities(time_step)

                # Update the momenta of the particles (accumulate scaling)
                self._update_particle_momenta(time_step, simulator.system)

                # Update thermostat positions (Only for debugging via conserved quantity)
                # self.t_positions = self.t_positions + 0.5 * time_step * self.t_velocities
                # self.t_positions_cell = self.t_positions_cell + 0.5 * time_step * self.t_velocities_cell

                # Recompute kinetic energy
                (
                    kinetic_energy_for_thermostat,
                    kinetic_energy_for_barostat,
                ) = self._compute_kinetic_energy(simulator.system)

                # Update the box velocities
                self._update_forces_barostat(
                    kinetic_energy_for_barostat, pressure, volume
                )
                self._update_box_velocities(time_step)

                # Update the thermostat force
                self._update_forces_thermostat(kinetic_energy_for_thermostat)

                # Update velocities and forces of remaining thermostats
                self._chain_backward(time_step)

        if self.detach:
            # self.t_positions = self.t_positions.detach()
            # self.t_positions_cell = self.t_positions_cell.detach()
            self.t_velocities = self.t_velocities.detach()
            self.t_velocities_cell = self.t_velocities_cell.detach()
            self.b_velocities_cell = self.b_velocities_cell.detach()

    def _init_kinetic_energy(self, system):
        """
        Auxiliary routine for initializing the kinetic energy and accumulating the scaling factor.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        self.scaling = 1.0
        # R x M x A x 3
        self.kinetic_energy = system.momenta ** 2 / system.masses * system.atom_masks

    def _compute_kinetic_energy(self, system):
        """
        Compute the current kinetic energy. Here an internal surrogate is used, which is scaled continuously.
        Since barostat and thermostat updates require different kinetic energy conventions in case of massive
        updates, two separate tensors are computed.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        Returns:
            torch.tensor: Current kinetic energy of the particles.
        """

        kinetic_energy_for_barostat = torch.sum(self.kinetic_energy, dim=[2, 3])

        if self.massive:
            kinetic_energy_for_thermostat = self.kinetic_energy
        else:
            kinetic_energy_for_thermostat = kinetic_energy_for_barostat[
                :, :, None, None
            ]

        return kinetic_energy_for_thermostat, kinetic_energy_for_barostat

    def _compute_kinetic_energy_cell(self):
        """
        Compute the kinetic energy of the cells.

        Returns:
            torch.tensor: Kinetic energy associated with the cells.
        """
        return self.b_masses_cell * self.b_velocities_cell ** 2

    def _compute_pressure(self, system):
        """
        Routine for computing the current pressure and volume associated with the simulated systems. The stress tensor
        is used for pressure computation.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.

        Returns:
            (torch.Tensor, torch.Tensor): Duple containing the pressures and volumes.
        """
        # Get the pressure (R x M)
        pressure = system.compute_pressure(kinetic_component=False, tensor=False)
        # Get the volume (R x M)
        volume = system.volume
        return pressure, volume

    def _chain_forward(self, time_step):
        """
        Forward propagation of the two Nose-Hoover chains attached to particles and cells.

        Args:
            time_step (float): Current timestep considering YS and multi-timestep integration.
        """
        # Update velocities of outermost bath
        self.t_velocities[..., -1] += 0.25 * self.t_forces[..., -1] * time_step
        self.t_velocities_cell[..., -1] += (
            0.25 * self.t_forces_cell[..., -1] * time_step
        )

        # Update the velocities moving through the beads of the chain
        for chain in range(self.chain_length - 2, -1, -1):
            t_coeff = torch.exp(-0.125 * time_step * self.t_velocities[..., chain + 1])
            b_coeff = torch.exp(
                -0.125 * time_step * self.t_velocities_cell[..., chain + 1]
            )

            self.t_velocities[..., chain] = (
                self.t_velocities[..., chain] * t_coeff ** 2
                + 0.25 * self.t_forces[..., chain] * t_coeff * time_step
            )
            self.t_velocities_cell[..., chain] = (
                self.t_velocities_cell[..., chain] * b_coeff ** 2
                + 0.25 * self.t_forces_cell[..., chain] * b_coeff * time_step
            )

    def _update_box_velocities(self, time_step):
        """
        Update the velocities of the additional degree of freedom associated with the simulation cells.

        Args:
            time_step (float): Current timestep considering YS and multi-timestep integration.
        """
        b_factor = torch.exp(-0.125 * time_step * self.t_velocities_cell[..., 0])
        self.b_velocities_cell = (
            b_factor ** 2 * self.b_velocities_cell
            + 0.25 * time_step * self.b_forces_cell * b_factor
        )

    def _update_particle_momenta(self, time_step, system):
        """
        Update the momenta of the particles, as well as the internal kinetic energy surrogate.

        Args:
            time_step (float): Current timestep considering YS and multi-timestep integration.
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        scaling = torch.exp(
            -0.5
            * time_step
            * (
                self.t_velocities[..., 0]
                + self.b_velocities_cell * (1 + 3 / self.degrees_of_freedom)
            )
        )
        system.momenta *= scaling * system.atom_masks
        self.kinetic_energy *= scaling ** 2

    def _chain_backward(self, time_step):
        """
        Backward propagation of the two Nose-Hoover chains attached to particles and cells. In addition, the repsective
        thermostat forces are updated.

        Args:
            time_step (float): Current timestep considering YS and multi-timestep integration.
        """
        # Update the thermostat velocities
        for chain in range(self.chain_length - 1):
            t_coeff = torch.exp(-0.125 * time_step * self.t_velocities[..., chain + 1])
            b_coeff = torch.exp(
                -0.125 * time_step * self.t_velocities_cell[..., chain + 1]
            )

            self.t_velocities[..., chain] = (
                self.t_velocities[..., chain] * t_coeff ** 2
                + 0.25 * self.t_forces[..., chain] * t_coeff * time_step
            )
            self.t_velocities_cell[..., chain] = (
                self.t_velocities_cell[..., chain] * b_coeff ** 2
                + 0.25 * self.t_forces_cell[..., chain] * b_coeff * time_step
            )

            # Update forces through chain
            self.t_forces[..., chain + 1] = (
                self.t_masses[..., chain] * self.t_velocities[..., chain] ** 2
                - self.kb_temperature
            ) / self.t_masses[..., chain + 1]
            self.t_forces_cell[..., chain + 1] = (
                self.t_masses_cell[..., chain] * self.t_velocities_cell[..., chain] ** 2
                - self.kb_temperature
            ) / self.t_masses_cell[..., chain + 1]

        # Update velocities of outermost thermostat
        self.t_velocities[..., -1] += 0.25 * self.t_forces[..., -1] * time_step
        self.t_velocities_cell[..., -1] += (
            0.25 * self.t_forces_cell[..., -1] * time_step
        )

    def _update_forces_thermostat(self, kinetic_energy):
        """
        Update the forces acting on the two innermost thermostats coupled to the particle and cell momenta.

        Args:
            kinetic_energy (torch.Tensor): Tensor containing the current kinetic energies of the systems.
        """
        # Compute forces on thermostat (R x M)
        self.t_forces[..., 0] = (
            kinetic_energy - self.degrees_of_freedom_particles * self.kb_temperature
        ) / self.t_masses[..., 0]

        # Get kinetic energy of barostat (R x M)
        kinetic_energy_cell = self._compute_kinetic_energy_cell()
        # Compute forces on cell thermostat
        self.t_forces_cell[..., 0] = (
            kinetic_energy_cell - self.kb_temperature
        ) / self.t_masses_cell[..., 0]

    def _update_forces_barostat(self, kinetic_energy, pressure, volume):
        """
        Update the forces acting on the barostat coupled to the cell.

        Args:
            kinetic_energy (torch.Tensor): Tensor containing the current kinetic energies of the systems.
            pressure (torch.Tensor): Current pressure of each system.
            volume (torch.Tensor): Current volume of each system.
        """
        self.b_forces_cell = (
            (1.0 + 3.0 / self.degrees_of_freedom) * kinetic_energy
            + 3.0 * volume * (pressure - self.target_pressure)
        ) / self.b_masses_cell

    def propagate_system(self, system):
        """
        Main routine for propagating the system positions and cells. Since this is modifed, no conventional velocity
        verlet integrator can be used.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        # Update the particle positions
        scaled_velocity = self.time_step * self.b_velocities_cell
        a_coeff = torch.exp(0.5 * scaled_velocity)[:, :, None, None]
        b_coeff = a_coeff * self.sinhdx.f(0.5 * scaled_velocity)[:, :, None, None]
        system.positions = (
            system.positions * a_coeff ** 2
            + system.momenta / system.masses * b_coeff * self.time_step
        )

        # Scale the cells
        cell_coeff = torch.exp(scaled_velocity)[:, :, None, None]
        system.cells = system.cells * cell_coeff

    # def compute_conserved(self, system):
    #    """
    #    Computed the conserved quantity. For debug purposes only.
    #    """
    #    conserved = (
    #            system.kinetic_energy[..., None, None]
    #            + system.energies[..., None, None]
    #            + 0.5 * torch.sum(self.t_velocities ** 2 * self.t_masses, 2)
    #            + 0.5 * torch.sum(self.t_velocities_cell ** 2 * self.t_masses_cell, 2)
    #            + 0.5 * self.b_velocities_cell ** 2 * self.b_masses_cell
    #            + self.degrees_of_freedom * self.kb_temperature * self.t_positions[..., 0]
    #            + self.kb_temperature * self.t_positions_cell[..., 0]
    #            + self.kb_temperature * torch.sum(self.t_positions[..., 1:], 2)
    #            + self.kb_temperature * torch.sum(self.t_positions_cell[..., 1:], 2)
    #            + self.target_pressure * system.volume
    #    )
    #    return conserved

    @property
    def state_dict(self):
        state_dict = {
            "chain_length": self.chain_length,
            "frequency": self.frequency,
            "cell_frequency": self.cell_frequency,
            "barostat_frequency": self.barostat_frequency,
            "kb_temperature": self.kb_temperature,
            "degrees_of_freedom": self.degrees_of_freedom,
            "t_masses": self.t_masses,
            "t_masses_cell": self.t_masses_cell,
            "b_masses_cell": self.b_masses_cell,
            "t_velocities": self.t_velocities,
            "t_velocities_cell": self.t_velocities_cell,
            "b_velocities_cell": self.b_velocities_cell,
            "t_forces": self.t_forces,
            "t_forces_cell": self.t_forces_cell,
            "b_forces_cell": self.b_forces_cell,
            "time_step": self.ys_time_step,
            "temperature_bath": self.temperature_bath,
            "target_pressure": self.target_pressure,
            "n_replicas": self.n_replicas,
            "multi_step": self.multi_step,
            "integration_order": self.integration_order,
            "massive": self.massive,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.chain_length = state_dict["chain_length"]
        self.frequency = state_dict["frequency"]
        self.cell_frequency = state_dict["cell_frequency"]
        self.barostat_frequency = state_dict["barostat_frequency"]
        self.kb_temperature = state_dict["kb_temperature"]
        self.degrees_of_freedom = state_dict["degrees_of_freedom"]
        self.t_masses = state_dict["t_masses"]
        self.t_masses_cell = state_dict["t_masses_cell"]
        self.b_masses_cell = state_dict["b_masses_cell"]
        self.t_velocities = state_dict["t_velocities"]
        self.t_velocities_cell = state_dict["t_velocities_cell"]
        self.b_velocities_cell = state_dict["b_velocities_cell"]
        self.t_forces = state_dict["t_forces"]
        self.t_forces_cell = state_dict["t_forces_cell"]
        self.b_forces_cell = state_dict["b_forces_cell"]
        self.ys_time_step = state_dict["time_step"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.target_pressure = state_dict["target_pressure"]
        self.n_replicas = state_dict["n_replicas"]
        self.multi_step = state_dict["multi_step"]
        self.integration_order = state_dict["integration_order"]
        self.massive = state_dict["massive"]

        self.initialized = True


class NHCBarostatAnisotropic(NHCBarostatIsotropic):
    """
    Parrinello--Rahman thermostat/barostat based on Nose-Hoover chains for fully anisotropic cell fluctuations. For a
    in-depth description of the algorithm, see[#nhc_barostat2]_. This barostat already contains a built in thermostat,
    so no further temperature control is necessary. As suggested in [#nhc_barostat2]_, two separate chains are used to
    thermostat particle and cell momenta. In order to propagate the system, operators in the form of matrix exponentials
    need to be applied, which requires several eigendecompositions of 3x3 matrices, making the barostat computationally
    intensive. In addition, accurate predictors of the stress tensor are required, since deformations of the cells are
    now possible.

    Args:
        target_pressure (float): Target pressure in bar.
        temperature_bath (float): Temperature of the external heat bath in Kelvin.
        time_constant (float): Thermostat time constant in fs
        time_constant_cell (float): Cell thermostat time constant in fs. If None is given (default), the same time
                                    constant as for the thermostat component is used.
        time_constant_barostat (float): Barostat time constant in fs. If None is given (default), the same time constant
                                        as for the thermostat component is used.
        chain_length (int): Number of Nose-Hoover thermostats applied in the chain.
        multi_step (int): Number of steps used for integrating the NH equations of motion (default=2)
        integration_order (int): Order of the Yoshida-Suzuki integrator used for propagating the thermostat (default=3).
        massive (bool): Apply individual thermostat chains to all particle degrees of freedom (default=False).
        detach (bool): Whether the computational graph should be detached after each simulation step. Default is true,
                       should be changed if differentiable MD is desired.

    References
    ----------
    .. [#nhc_barostat2] Martyna, Tuckerman, Tobias, Klein:
       Explicit reversible integrators for extended systems dynamics.
       Molecular Physics, 87(5), 1117-1157. 1996.
    """

    def __init__(
        self,
        target_pressure,
        temperature_bath,
        time_constant,
        time_constant_cell=None,
        time_constant_barostat=None,
        chain_length=4,
        multi_step=4,
        integration_order=7,
        massive=False,
        detach=True,
    ):
        super(NHCBarostatAnisotropic, self).__init__(
            target_pressure=target_pressure,
            temperature_bath=temperature_bath,
            time_constant=time_constant,
            time_constant_cell=time_constant_cell,
            time_constant_barostat=time_constant_barostat,
            chain_length=chain_length,
            multi_step=multi_step,
            integration_order=integration_order,
            massive=massive,
            detach=detach,
        )

    def _init_barostat_variables(self):
        """
        Initialize all quantities required for the barostat component.
        """
        # Set barostat masses
        self.b_masses_cell = torch.ones(
            self.n_replicas, self.n_molecules, 1, 1, device=self.device
        )
        self.b_masses_cell[..., 0, 0] = (
            (self.degrees_of_freedom + 3)
            * self.kb_temperature
            / self.barostat_frequency ** 2
        ) / 3.0

        # Remaining barostat variables
        self.b_velocities_cell = torch.zeros(
            self.n_replicas, self.n_molecules, 3, 3, device=self.device
        )
        self.b_forces_cell = torch.zeros_like(
            self.b_velocities_cell, device=self.device
        )

        # Auxiliary identity matrix for broadcasting
        self.aux_eye = torch.eye(3, device=self.device)[None, None, :, :]

    def _init_kinetic_energy(self, system):
        """
        This routine is no longer required, since it is no longer possible to accumulate the barostat
        action.

        Args:
           system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        pass

    def _compute_kinetic_energy(self, system):
        """
        Compute the current kinetic energy tensor.
        Since barostat and thermostat updates require different kinetic energy conventions in case of massive
        updates, two separate tensors are computed.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        Returns:
            torch.tensor: Current kinetic energy tensor of the particles.
        """
        # Here we need the full tensor (R x M x 3 x 3)
        # Kinetic energy can be computed as Tr[Etens]
        kinetic_energy_tensor = 2.0 * system.kinetic_energy_tensor

        if self.massive:
            kinetic_energy_for_thermostat = system.momenta ** 2 / system.masses
        else:
            kinetic_energy_for_thermostat = torch.einsum(
                "abii->ab", kinetic_energy_tensor
            )[:, :, None, None]

        return kinetic_energy_for_thermostat, kinetic_energy_tensor

    def _compute_kinetic_energy_cell(self):
        """
        Compute the kinetic energy of the cells.

        Returns:
            torch.tensor: Kinetic energy associated with the cells.
        """
        b_cell_sq = torch.matmul(
            self.b_velocities_cell.transpose(2, 3), self.b_velocities_cell
        )
        # Einsum computes the trace
        return (
            self.b_masses_cell * torch.einsum("abii->ab", b_cell_sq)[:, :, None, None]
        )

    def _compute_pressure(self, system):
        """
        Routine for computing the current pressure tensor and volume associated with the simulated systems.
        The pressure tensor is equivalent to the negative stress tensor.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.

        Returns:
            (torch.Tensor, torch.Tensor): Duple containing the pressure tensors and volumes.
        """
        # Get the pressure (R x M x 3 x 3)
        pressure = system.compute_pressure(kinetic_component=False, tensor=True)
        # Get the volume (R x M x 1 x 1)
        volume = system.volume[..., None]
        return pressure, volume

    def _update_particle_momenta(self, time_step, system):
        """
        Update the momenta of the particles. In contrast to the isotropic case, an eigenvalue problem needs to be
        solved.

        Args:
            time_step (float): Current timestep considering YS and multi-timestep integration.
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        # Compute auxiliary velocity tensor for propagation
        # vtemp = (
        #         self.b_velocities_cell
        #         + (
        #                 torch.einsum("abii->ab", self.b_velocities_cell)[:, :, None, None]
        #                 / self.degrees_of_freedom
        #                 + self.t_velocities[..., 0]
        #         ) * self.aux_eye
        # )

        # Compute eigenvectors and values for matrix exponential operator
        # eigval -> (R x M x 3)
        # eigvec -> (R x M x 3 x 3)
        # eigval, eigvec = torch.symeig(vtemp, eigenvectors=True)
        # operator = torch.exp(-0.5 * eigval * self.time_step)[:, :, None, :]

        # Since the original matrix consist of matrix + diagonal term, the operator can be split
        # in order to apply massive particle thermostats
        # TODO: check sinh in momenta updates as in Tuckerman book
        vtemp = (
            self.b_velocities_cell
            + (
                torch.einsum("abii->ab", self.b_velocities_cell)
                / self.degrees_of_freedom
            )[:, :, None, None]
            * self.aux_eye
        )
        eigval, eigvec = torch.symeig(vtemp, eigenvectors=True)
        operator1 = torch.exp(-0.5 * eigval * time_step)[:, :, None, :]
        operator2 = torch.exp(-0.5 * self.t_velocities[..., 0] * time_step)

        # The following procedure computes the matrix exponential of vtemp and applies it to
        # the momenta.
        # p' = p * c
        momenta_tmp = torch.matmul(system.momenta / system.masses, eigvec)
        # Multiply by operator
        momenta_tmp = momenta_tmp * operator1
        # Transform back
        # p = p' * c.T
        system.momenta = (
            torch.matmul(momenta_tmp, eigvec.transpose(2, 3))
            * operator2
            * system.masses
            * system.atom_masks
        )

    def _update_forces_thermostat(self, kinetic_energy):
        """
        Update the forces acting on the two innermost thermostats coupled to the particle and cell momenta.
        The standard kinetic energy is computed as the trace of the kinetic energy tensor.

        Args:
            kinetic_energy (torch.Tensor): Tensor containing the current kinetic energies of the systems.
        """
        # Compute Ekin from tensor
        # kinetic_energy = torch.einsum("abii->ab", kinetic_energy)
        # Compute forces on thermostat (R x M)
        self.t_forces[..., 0] = (
            kinetic_energy - self.degrees_of_freedom_particles * self.kb_temperature
        ) / self.t_masses[..., 0]

        # Get kinetic energy of barostat (R x M)
        kinetic_energy_cell = self._compute_kinetic_energy_cell()
        # Compute forces on cell thermostat
        self.t_forces_cell[..., 0] = (
            kinetic_energy_cell - 9.0 * self.kb_temperature
        ) / self.t_masses_cell[..., 0]

    def _update_forces_barostat(self, kinetic_energy, pressure, volume):
        """
        Update the forces acting on the barostat coupled to the cell.
        The standard kinetic energy is computed as the trace of the kinetic energy tensor.

        Args:
            kinetic_energy (torch.Tensor): Tensor containing the current kinetic energies of the systems.
            pressure (torch.Tensor): Current pressure of each system.
            volume (torch.Tensor): Current volume of each system.
        """
        kinetic_energy_scalar = torch.einsum("abii->ab", kinetic_energy)[
            :, :, None, None
        ]
        self.b_forces_cell = (
            kinetic_energy_scalar
            / self.degrees_of_freedom[:, :, None, None]
            * self.aux_eye
            + kinetic_energy
            + volume * (pressure - self.aux_eye * self.target_pressure)
        ) / self.b_masses_cell

    def propagate_system(self, system):
        """
        Main routine for propagating the system positions and cells. Compared to the standard velocity verlet, this
        routine is heavily modified and makes use of eigendecomposition.
        verlet integrator can be used.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        # Compute eigenvectors and values for matrix exponential operator
        # eigval -> (R x M x 3)
        # eigvec -> (R x M x 3 x 3)
        eigval, eigvec = torch.symeig(self.b_velocities_cell, eigenvectors=True)

        evaldt2 = 0.5 * eigval[:, :, None, :] * self.time_step

        # Compute exponential operator and sinh(x)/x operator (approximated)
        a_coeff = torch.exp(evaldt2)
        b_coeff = a_coeff * self.sinhdx.f(evaldt2)

        # Transform positons, velocities and cells via the eigenvectors
        positions_tmp = torch.matmul(system.positions, eigvec)
        velocities_tmp = torch.matmul(system.momenta / system.masses, eigvec)
        cells_tmp = torch.matmul(system.cells, eigvec)

        # Apply the propagator to the positions
        positions_tmp = (
            positions_tmp * a_coeff ** 2 + velocities_tmp * b_coeff * self.time_step
        )

        # Apply the propagator to the cells
        cells_tmp = cells_tmp * a_coeff ** 2

        # Transform everything back and update the system
        system.positions = torch.matmul(positions_tmp, eigvec.transpose(2, 3))
        system.cells = torch.matmul(cells_tmp, eigvec.transpose(2, 3))

    # def compute_conserved(self, system):
    #    """
    #    Computed the conserved quantity. For debug purposes only.
    #    """
    #    conserved = (
    #            system.kinetic_energy[..., None, None]
    #            + system.energies[..., None, None]
    #            + 0.5 * torch.sum(self.t_velocities ** 2 * self.t_masses, 2)
    #            + 0.5 * torch.sum(self.t_velocities_cell ** 2 * self.t_masses_cell, 2)
    #            + 0.5 * self._compute_kinetic_energy_cell()
    #            + self.degrees_of_freedom * self.kb_temperature * self.t_positions[..., 0]
    #            + 9.0 * self.kb_temperature * self.t_positions_cell[..., 0]
    #            + self.kb_temperature * torch.sum(self.t_positions[..., 1:], 2)
    #            + self.kb_temperature * torch.sum(self.t_positions_cell[..., 1:], 2)
    #            + self.target_pressure * system.volume
    #    )
    #    return conserved

    @property
    def state_dict(self):
        state_dict = {
            "chain_length": self.chain_length,
            "frequency": self.frequency,
            "cell_frequency": self.cell_frequency,
            "barostat_frequency": self.barostat_frequency,
            "kb_temperature": self.kb_temperature,
            "degrees_of_freedom": self.degrees_of_freedom,
            "t_masses": self.t_masses,
            "t_masses_cell": self.t_masses_cell,
            "b_masses_cell": self.b_masses_cell,
            "t_velocities": self.t_velocities,
            "t_velocities_cell": self.t_velocities_cell,
            "b_velocities_cell": self.b_velocities_cell,
            "t_forces": self.t_forces,
            "t_forces_cell": self.t_forces_cell,
            "b_forces_cell": self.b_forces_cell,
            "time_step": self.ys_time_step,
            "temperature_bath": self.temperature_bath,
            "target_pressure": self.target_pressure,
            "n_replicas": self.n_replicas,
            "multi_step": self.multi_step,
            "integration_order": self.integration_order,
            "aux_eye": self.aux_eye,
            "massive": self.massive,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.chain_length = state_dict["chain_length"]
        self.frequency = state_dict["frequency"]
        self.cell_frequency = state_dict["cell_frequency"]
        self.barostat_frequency = state_dict["barostat_frequency"]
        self.kb_temperature = state_dict["kb_temperature"]
        self.degrees_of_freedom = state_dict["degrees_of_freedom"]
        self.t_masses = state_dict["t_masses"]
        self.t_masses_cell = state_dict["t_masses_cell"]
        self.b_masses_cell = state_dict["b_masses_cell"]
        self.t_velocities = state_dict["t_velocities"]
        self.t_velocities_cell = state_dict["t_velocities_cell"]
        self.b_velocities_cell = state_dict["b_velocities_cell"]
        self.t_forces = state_dict["t_forces"]
        self.t_forces_cell = state_dict["t_forces_cell"]
        self.b_forces_cell = state_dict["b_forces_cell"]
        self.ys_time_step = state_dict["time_step"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.target_pressure = state_dict["target_pressure"]
        self.n_replicas = state_dict["n_replicas"]
        self.multi_step = state_dict["multi_step"]
        self.integration_order = state_dict["integration_order"]
        self.aux_eye = state_dict["aux_eye"]
        self.massive = state_dict["massive"]

        self.initialized = True


class RPMDBarostat(BarostatHook):
    """
    Barostat for ring polymer molecular dynamics simulations. This barostat is based on the algorithm introduced in
    [#rpmd_barostat1]_ and uses a PILE thermostat for the cell. It can be combined with any RPMD thermostat for
    temperature control. The barostat only acts on the centroid mode of the ring polymer and is designed for isotropic
    cell fluctuations.

    Args:
        target_pressure (float): Target pressure of the system (in bar).
        temperature_bath (float): Target temperature applied to the cell fluctuations.
        detach (bool): Whether the computational graph should be detached after each simulation step. Default is true,
                       should be changed if differentiable MD is desired.

    References
    ----------
    .. [#rpmd_barostat1] Kapil, et al.
                         i-PI 2.0: A universal force engine for advanced molecular simulations.
                         Computer Physics Communications, 236, 214-223. 2019.
    """

    def __init__(self, target_pressure, temperature_bath, time_constant, detach=True):
        super(RPMDBarostat, self).__init__(
            target_pressure=target_pressure,
            temperature_bath=temperature_bath,
            detach=detach,
        )
        self.frequency = 1.0 / (time_constant * MDUnits.fs2internal)
        self.kb_temperature = temperature_bath * MDUnits.kB
        self.transformation = None
        self.propagator = None
        self.cell_momenta = None
        self.sinhdx = StableSinhDiv()

    def _init_barostat(self, simulator):
        """
        Initialize the thermostat coefficients and barostat quantities.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                    the time step, system, etc.
        """
        # Get normal mode transformer and propagator for position and cell update
        self.transformation = simulator.integrator.transformation
        self.propagator = simulator.integrator.propagator

        # Set up centroid momenta of cell (one for every molecule)
        self.cell_momenta = torch.zeros(
            self.n_molecules, device=simulator.system.device
        )
        self.mass = (
            3.0 * simulator.system.n_atoms / self.frequency ** 2 * self.kb_temperature
        )

        # Set up cell thermostat coefficients
        self.c1 = torch.exp(
            -0.5 * torch.ones(1, device=self.device) * self.frequency * self.time_step
        )
        self.c2 = torch.sqrt(
            self.n_replicas * self.mass * self.kb_temperature * (1.0 - self.c1 ** 2)
        )

    def _apply_barostat(self, simulator):
        """
        Apply the thermostat. This simply propagates the cell momenta under the influence of a PILE thermostat.

        Args:
            simulator (schnetpack.simulation_hooks.simulator.Simulator): Main simulator class containing information on
                                                                the time step, system, etc.
        """
        # Propagate cell momenta during half-step
        self.cell_momenta = self.c1 * self.cell_momenta + self.c2 * torch.randn_like(
            self.cell_momenta
        )
        if self.detach:
            self.cell_momenta = self.cell_momenta.detach()

    def propagate_system(self, system):
        """
        Main routine for propagating the ring polymer and the cells. The barostat acts only on the centroid, while the
        remaining replicas are propagated in the conventional way.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        # Transform to normal mode representation
        positions_normal = self.transformation.beads2normal(system.positions)
        momenta_normal = self.transformation.beads2normal(system.momenta)

        # Propagate centroid mode of the ring polymer
        reduced_momenta = (self.cell_momenta / self.mass)[:, None, None]
        scaling = torch.exp(-self.time_step * reduced_momenta)

        momenta_normal[0] = momenta_normal[0] * scaling
        positions_normal[0] = (
            positions_normal[0] / scaling
            + self.sinhdx.f(self.time_step * reduced_momenta)
            * (momenta_normal[0] / system.masses[0])
            * self.time_step
        )

        # Update cells
        system.cells = system.cells / scaling[None, ...]

        # Propagate the remaining modes of the ring polymer
        momenta_normal[1:] = (
            self.propagator[1:, 0, 0] * momenta_normal[1:]
            + self.propagator[1:, 0, 1] * positions_normal[1:] * system.masses
        )
        positions_normal[1:] = (
            self.propagator[1:, 1, 0] * momenta_normal[1:] / system.masses
            + self.propagator[1:, 1, 1] * positions_normal[1:]
        )

        # Transform back to bead representation
        system.positions = self.transformation.normal2beads(positions_normal)
        system.momenta = self.transformation.normal2beads(momenta_normal)

    def propagate_barostat_half_step(self, system):
        """
        Propagate the momenta of the thermostat attached to the barostat during each half-step.

        Args:
            system (schnetpack.md.System): System class containing all molecules and their
                             replicas.
        """
        centroid_momenta = self.transformation.beads2normal(system.momenta)[0]
        centroid_forces = self.transformation.beads2normal(system.forces)[0]

        # Compute pressure component (volume[0] can be used, since the volume is scaled equally for all replicas)
        component_1 = (
            3.0
            * self.n_replicas
            * (
                system.volume[0]
                * (
                    torch.mean(system.compute_pressure(kinetic_component=True), dim=0)
                    - self.target_pressure
                )
                + self.kb_temperature
            )
        )

        # Compute components based on forces and momenta
        force_by_mass = centroid_forces / system.masses[0]

        component_2 = torch.sum(force_by_mass * centroid_momenta, dim=[1, 2])
        component_3 = torch.sum(force_by_mass * centroid_forces / 3, dim=[1, 2])

        # Update cell momenta
        self.cell_momenta = (
            self.cell_momenta
            + (0.5 * self.time_step) * component_1
            + (0.5 * self.time_step) ** 2 * component_2
            + (0.5 * self.time_step) ** 3 * component_3
        )

        if self.detach:
            self.cell_momenta = self.cell_momenta.detach()

    @property
    def state_dict(self):
        state_dict = {
            "frequency": self.frequency,
            "kb_temperature": self.kb_temperature,
            "transformation": self.transformation,
            "propagator": self.propagator,
            "cell_momenta": self.cell_momenta,
            "mass": self.mass,
            "c1": self.c1,
            "c2": self.c2,
            "temperature_bath": self.temperature_bath,
            "target_pressure": self.target_pressure,
            "n_replicas": self.n_replicas,
        }
        return state_dict

    @state_dict.setter
    def state_dict(self, state_dict):
        self.frequency = state_dict["frequency"]
        self.kb_temperature = state_dict["kb_temperature"]
        self.transformation = state_dict["transformation"]
        self.propagator = state_dict["propagator"]
        self.cell_momenta = state_dict["cell_momenta"]
        self.mass = state_dict["mass"]
        self.c1 = state_dict["c1"]
        self.c2 = state_dict["c2"]
        self.temperature_bath = state_dict["temperature_bath"]
        self.target_pressure = state_dict["target_pressure"]
        self.n_replicas = state_dict["n_replicas"]

        self.initialized = True



import os
import logging
from shutil import rmtree
import schnetpack as spk


__all__ = ["setup_run"]


def setup_run(args):
    argparse_dict = vars(args)
    jsonpath = os.path.join(args.modelpath, "args.json")
    if args.mode == "train":

        # build modeldir
        if args.overwrite and os.path.exists(args.modelpath):
            logging.info("existing model will be overwritten...")
            rmtree(args.modelpath)
        if not os.path.exists(args.modelpath):
            os.makedirs(args.modelpath)

        # store training arguments
        spk.utils.spk_utils.to_json(jsonpath, argparse_dict)

        spk.utils.spk_utils.set_random_seed(args.seed)
        train_args = args
    else:
        # check if modelpath is valid
        if not os.path.exists(args.modelpath):
            raise spk.utils.ScriptError(
                "The selected modeldir does not exist " "at {}!".format(args.modelpath)
            )

        # load training arguments
        train_args = spk.utils.spk_utils.read_from_json(jsonpath)

    # apply alias definitions
    train_args = apply_aliases(train_args)
    return train_args


def apply_aliases(args):
    # force alias for custom dataset
    if args.dataset == "custom":
        if args.force is not None:
            if args.derivative is not None:
                raise spk.utils.ScriptError(
                    "Force and derivative define the same property. Please don`t use "
                    "both."
                )
            args.derivative = args.force
            args.negative_dr = True

            # add rho value if selected
            if "force" in args.rho.keys():
                args.rho["derivative"] = args.rho.pop("force")

    return args



import logging
import os
import shutil
import tarfile
import tempfile
from urllib import request as request

import h5py
import numpy as np
from ase import Atoms
from ase.units import Hartree

import schnetpack as spk
from schnetpack.datasets import DownloadableAtomsData


class ANI1(DownloadableAtomsData):
    """ANI1 benchmark database.

    This class adds convenience functions to download ANI1 from figshare and
    load the data into pytorch.

    Args:

        dbpath (str): path to directory containing database.
        download (bool, optional): enable downloading if database does not exists.
        subset (list, optional): Deprecated! Do not use! Subsets are created with
            AtomsDataSubset class.
        load_only (list, optional): reduced set of properties to be loaded
        collect_triples (bool, optional): Set to True if angular features are needed.
        num_heavy_atoms (int, optional): number of heavy atoms.
            (See 'Table 1' in Ref. [#ani1]_)
        high_energies (bool, optional): add high energy conformations.
            (See 'Technical Validation' of Ref. [#ani1]_)
        environment_provider (spk.environment.BaseEnvironmentProvider): define how
            neighborhood is calculated
            (default=spk.environment.SimpleEnvironmentProvider).

    References:
        .. [#ani1] https://arxiv.org/abs/1708.04987

    """

    # properties:
    energy = "energy"

    reference = {energy: 0}

    self_energies = {
        "H": -0.500607632585,
        "C": -37.8302333826,
        "N": -54.5680045287,
        "O": -75.0362229210,
    }

    def __init__(
        self,
        dbpath,
        download=True,
        subset=None,
        load_only=None,
        collect_triples=False,
        num_heavy_atoms=8,
        high_energies=False,
        environment_provider=spk.environment.SimpleEnvironmentProvider(),
    ):
        available_properties = [ANI1.energy]
        units = [Hartree]

        self.num_heavy_atoms = num_heavy_atoms
        self.high_energies = high_energies

        super().__init__(
            dbpath=dbpath,
            subset=subset,
            download=download,
            load_only=load_only,
            collect_triples=collect_triples,
            available_properties=available_properties,
            units=units,
            environment_provider=environment_provider,
        )

    def _download(self):
        self._load_data()

        atref, labels = self._create_atoms_ref()

        self.set_metadata({"atomrefs": atref.tolist(), "atref_labels": labels})

    def _load_data(self):
        logging.info("downloading ANI-1 data...")
        tmpdir = tempfile.mkdtemp("ani1")
        tar_path = os.path.join(tmpdir, "ANI1_release.tar.gz")
        raw_path = os.path.join(tmpdir, "data")
        url = "https://ndownloader.figshare.com/files/9057631"

        request.urlretrieve(url, tar_path)
        logging.info("Done.")

        tar = tarfile.open(tar_path)
        tar.extractall(raw_path)
        tar.close()

        logging.info("parse files...")
        for i in range(1, self.num_heavy_atoms + 1):
            file_name = os.path.join(raw_path, "ANI-1_release", "ani_gdb_s0%d.h5" % i)
            logging.info("start to parse %s" % file_name)
            self._load_h5_file(file_name)

        logging.info("done...")

        shutil.rmtree(tmpdir)

    def _load_h5_file(self, file_name):
        atoms_list = []
        properties_list = []

        store = h5py.File(file_name)
        for file_key in store:
            for molecule_key in store[file_key]:
                molecule_group = store[file_key][molecule_key]
                species = "".join([str(s)[-2] for s in molecule_group["species"]])
                positions = molecule_group["coordinates"]
                energies = molecule_group["energies"]

                # loop over conformations
                for i in range(energies.shape[0]):
                    atm = Atoms(species, positions[i])
                    energy = energies[i] * self.units[self.energy]
                    properties = {self.energy: energy}
                    atoms_list.append(atm)
                    properties_list.append(properties)

                # high energy conformations as described in 'Technical Validation'
                # section of https://arxiv.org/abs/1708.04987
                if self.high_energies:
                    high_energy_positions = molecule_group["coordinatesHE"]
                    high_energies = molecule_group["energiesHE"]

                    # loop over high energy conformations
                    for i in range(high_energies.shape[0]):
                        atm = Atoms(species, high_energy_positions[i])
                        high_energy = high_energies[i] * self.units[self.high_energies]
                        properties = {self.energy: high_energy}
                        atoms_list.append(atm)
                        properties_list.append(properties)

        # write data to ase db
        self.add_systems(atoms_list, property_list=properties_list)

    def _create_atoms_ref(self):
        atref = np.zeros((100, 6))
        labels = self.load_only

        # converts units to eV (which are set to one in ase)
        atref[1, :] = self.self_energies["H"] * self.units["energy"]
        atref[6, :] = self.self_energies["C"] * self.units["energy"]
        atref[7, :] = self.self_energies["N"] * self.units["energy"]
        atref[8, :] = self.self_energies["O"] * self.units["energy"]

        return atref, labels



import torch
from torch import nn


def atom_distances(
    positions,
    neighbors,
    cell=None,
    cell_offsets=None,
    return_vecs=False,
    normalize_vecs=False,
    neighbor_mask=None,
):
    r"""Compute distance of every atom to its neighbors.

    This function uses advanced torch indexing to compute differentiable distances
    of every central atom to its relevant neighbors.

    Args:
        positions (torch.Tensor):
            atomic Cartesian coordinates with (N_b x N_at x 3) shape
        neighbors (torch.Tensor):
            indices of neighboring atoms to consider with (N_b x N_at x N_nbh) shape
        cell (torch.tensor, optional):
            periodic cell of (N_b x 3 x 3) shape
        cell_offsets (torch.Tensor, optional) :
            offset of atom in cell coordinates with (N_b x N_at x N_nbh x 3) shape
        return_vecs (bool, optional): if True, also returns direction vectors.
        normalize_vecs (bool, optional): if True, normalize direction vectors.
        neighbor_mask (torch.Tensor, optional): boolean mask for neighbor positions.

    Returns:
        (torch.Tensor, torch.Tensor):
            distances:
                distance of every atom to its neighbors with
                (N_b x N_at x N_nbh) shape.

            dist_vec:
                direction cosines of every atom to its
                neighbors with (N_b x N_at x N_nbh x 3) shape (optional).

    """

    # Construct auxiliary index vector
    n_batch = positions.size()[0]
    idx_m = torch.arange(n_batch, device=positions.device, dtype=torch.long)[
        :, None, None
    ]
    # Get atomic positions of all neighboring indices
    pos_xyz = positions[idx_m, neighbors[:, :, :], :]

    # Subtract positions of central atoms to get distance vectors
    dist_vec = pos_xyz - positions[:, :, None, :]

    # add cell offset
    if cell is not None:
        B, A, N, D = cell_offsets.size()
        cell_offsets = cell_offsets.view(B, A * N, D)
        offsets = cell_offsets.bmm(cell)
        offsets = offsets.view(B, A, N, D)
        dist_vec += offsets

    # Compute vector lengths
    distances = torch.norm(dist_vec, 2, 3)

    if neighbor_mask is not None:
        # Avoid problems with zero distances in forces (instability of square
        # root derivative at 0) This way is neccessary, as gradients do not
        # work with inplace operations, such as e.g.
        # -> distances[mask==0] = 0.0
        tmp_distances = torch.zeros_like(distances)
        tmp_distances[neighbor_mask != 0] = distances[neighbor_mask != 0]
        distances = tmp_distances

    if return_vecs:
        tmp_distances = torch.ones_like(distances)
        tmp_distances[neighbor_mask != 0] = distances[neighbor_mask != 0]

        if normalize_vecs:
            dist_vec = dist_vec / tmp_distances[:, :, :, None]
        return distances, dist_vec
    return distances


class AtomDistances(nn.Module):
    r"""Layer for computing distance of every atom to its neighbors.

    Args:
        return_directions (bool, optional): if True, the `forward` method also returns
            normalized direction vectors.

    """

    def __init__(self, return_directions=False):
        super(AtomDistances, self).__init__()
        self.return_directions = return_directions

    def forward(
        self, positions, neighbors, cell=None, cell_offsets=None, neighbor_mask=None
    ):
        r"""Compute distance of every atom to its neighbors.

        Args:
            positions (torch.Tensor): atomic Cartesian coordinates with
                (N_b x N_at x 3) shape.
            neighbors (torch.Tensor): indices of neighboring atoms to consider
                with (N_b x N_at x N_nbh) shape.
            cell (torch.tensor, optional): periodic cell of (N_b x 3 x 3) shape.
            cell_offsets (torch.Tensor, optional): offset of atom in cell coordinates
                with (N_b x N_at x N_nbh x 3) shape.
            neighbor_mask (torch.Tensor, optional): boolean mask for neighbor
                positions. Required for the stable computation of forces in
                molecules with different sizes.

        Returns:
            torch.Tensor: layer output of (N_b x N_at x N_nbh) shape.

        """
        return atom_distances(
            positions,
            neighbors,
            cell,
            cell_offsets,
            return_vecs=self.return_directions,
            normalize_vecs=True,
            neighbor_mask=neighbor_mask,
        )


def triple_distances(
    positions,
    neighbors_j,
    neighbors_k,
    offset_idx_j=None,
    offset_idx_k=None,
    cell=None,
    cell_offsets=None,
):
    """
    Get all distances between atoms forming a triangle with the central atoms.
    Required e.g. for angular symmetry functions.

    Args:
        positions (torch.Tensor): Atomic positions
        neighbors_j (torch.Tensor): Indices of first neighbor in triangle
        neighbors_k (torch.Tensor): Indices of second neighbor in triangle
        offset_idx_j (torch.Tensor): Indices for offets of neighbors j (for PBC)
        offset_idx_k (torch.Tensor): Indices for offets of neighbors k (for PBC)
        cell (torch.tensor, optional): periodic cell of (N_b x 3 x 3) shape.
        cell_offsets (torch.Tensor, optional): offset of atom in cell coordinates
            with (N_b x N_at x N_nbh x 3) shape.

    Returns:
        torch.Tensor: Distance between central atom and neighbor j
        torch.Tensor: Distance between central atom and neighbor k
        torch.Tensor: Distance between neighbors

    """
    nbatch, _, _ = neighbors_k.size()
    idx_m = torch.arange(nbatch, device=positions.device, dtype=torch.long)[
        :, None, None
    ]

    pos_j = positions[idx_m, neighbors_j[:], :]
    pos_k = positions[idx_m, neighbors_k[:], :]

    if cell is not None:
        # Get the offsets into true cartesian values
        B, A, N, D = cell_offsets.size()

        cell_offsets = cell_offsets.view(B, A * N, D)
        offsets = cell_offsets.bmm(cell)
        offsets = offsets.view(B, A, N, D)

        # Get the offset values for j and k atoms
        B, A, T = offset_idx_j.size()

        # Collapse batch and atoms position for easier indexing
        offset_idx_j = offset_idx_j.view(B * A, T)
        offset_idx_k = offset_idx_k.view(B * A, T)
        offsets = offsets.view(B * A, -1, D)

        # Construct auxiliary aray for advanced indexing
        idx_offset_m = torch.arange(B * A, device=positions.device, dtype=torch.long)[
            :, None
        ]

        # Restore proper dmensions
        offset_j = offsets[idx_offset_m, offset_idx_j[:]].view(B, A, T, D)
        offset_k = offsets[idx_offset_m, offset_idx_k[:]].view(B, A, T, D)

        # Add offsets
        pos_j = pos_j + offset_j
        pos_k = pos_k + offset_k

    # if positions.is_cuda:
    #    idx_m = idx_m.pin_memory().cuda(async=True)

    # Get the real positions of j and k
    R_ij = pos_j - positions[:, :, None, :]
    R_ik = pos_k - positions[:, :, None, :]
    R_jk = pos_j - pos_k

    # + 1e-9 to avoid division by zero
    r_ij = torch.norm(R_ij, 2, 3) + 1e-9
    r_ik = torch.norm(R_ik, 2, 3) + 1e-9
    r_jk = torch.norm(R_jk, 2, 3) + 1e-9

    return r_ij, r_ik, r_jk


class TriplesDistances(nn.Module):
    """
    Layer that gets all distances between atoms forming a triangle with the
    central atoms. Required e.g. for angular symmetry functions.
    """

    def __init__(self):
        super(TriplesDistances, self).__init__()

    def forward(self, positions, neighbors_j, neighbors_k):
        """
        Args:
            positions (torch.Tensor): Atomic positions
            neighbors_j (torch.Tensor): Indices of first neighbor in triangle
            neighbors_k (torch.Tensor): Indices of second neighbor in triangle

        Returns:
            torch.Tensor: Distance between central atom and neighbor j
            torch.Tensor: Distance between central atom and neighbor k
            torch.Tensor: Distance between neighbors

        """
        return triple_distances(positions, neighbors_j, neighbors_k)


def neighbor_elements(atomic_numbers, neighbors):
    """
    Return the atomic numbers associated with the neighboring atoms. Can also
    be used to gather other properties by neighbors if different atom-wise
    Tensor is passed instead of atomic_numbers.

    Args:
        atomic_numbers (torch.Tensor): Atomic numbers (Nbatch x Nat x 1)
        neighbors (torch.Tensor): Neighbor indices (Nbatch x Nat x Nneigh)

    Returns:
        torch.Tensor: Atomic numbers of neighbors (Nbatch x Nat x Nneigh)

    """
    # Get molecules in batch
    n_batch = atomic_numbers.size()[0]
    # Construct auxiliary index
    idx_m = torch.arange(n_batch, device=atomic_numbers.device, dtype=torch.long)[
        :, None, None
    ]
    # Get neighbors via advanced indexing
    neighbor_numbers = atomic_numbers[idx_m, neighbors[:, :, :]]
    return neighbor_numbers


class NeighborElements(nn.Module):
    """
    Layer to obtain the atomic numbers associated with the neighboring atoms.
    """

    def __init__(self):
        super(NeighborElements, self).__init__()

    def forward(self, atomic_numbers, neighbors):
        """
        Args:
            atomic_numbers (torch.Tensor): Atomic numbers (Nbatch x Nat x 1)
            neighbors (torch.Tensor): Neighbor indices (Nbatch x Nat x Nneigh)

        Returns:
            torch.Tensor: Atomic numbers of neighbors (Nbatch x Nat x Nneigh)
        """
        return neighbor_elements(atomic_numbers, neighbors)



import numpy as np
import torch
from ase import units

from schnetpack.md.utils import MDUnits


# __all__ = [
#     "load_gle_matrices",
#     "GLEMatrixParser",
#     "YSWeights"
# ]


def load_gle_matrices(filename):
    """
    Load GLE thermostat files formatted in raw format as generated via http://gle4md.org/index.html?page=matrix
    The generated matrices are torch tensors of the shape normal_modes x s+1 x s+1, where normal_modes is 1 except in
    the case of the PIGLET thermostat and s is the number of degrees of freedom added via GLE. Automatically recognizes
    used units and converts them to atomic units.

    Args:
        filename (str): Path to the file the GLE thermostat parameters should be loaded from.

    Returns:
        tuple: Tuple of two square torch tensors containing the a_matrix and c_matrix parameters required to
               initialize GLE type thermostats.
    """
    a_matrix = GLEMatrixParser(
        "A MATRIX:", stop="C MATRIX:", split="Matrix for normal mode"
    )
    c_matrix = GLEMatrixParser("C MATRIX:", split="Matrix for normal mode")

    try:
        with open(filename) as glefile:
            for line in glefile:
                a_matrix.read_line(line)
                c_matrix.read_line(line)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Could not open {:s} for reading. Please use GLE parameter files "
            "generated via http://gle4md.org/index.html?page=matrix".format(filename)
        )

    return a_matrix.matrix, c_matrix.matrix


class GLEMatrixParser:
    """
    General parser for GLE thermostat files. Reads from start string until end of file or a given stop string. If the
    argument split is specified, the read matrices are split at the given token.  Automatically recognizes
    used units and converts them to atomic units.

    Args:
        start (str): Token when to start reading.
        stop (str): Token when to stop reading. If None (default) reads until eno of file.
        split (str): If the given token is encountered, matrices are split at this point. If None (default), no split is
                     performed.
    """

    # Automatically recognized format and converts to units
    # TODO: think about this carefully
    unit_conversions = {
        "atomic time units^-1": 1.0 / MDUnits.unit2internal("aut"),
        "seconds^-1": 1.0 / MDUnits.unit2internal("s"),
        "femtoseconds^-1": 1.0 / MDUnits.unit2internal("fs"),
        "picoseconds^-1": 1.0 / 1000.0 / MDUnits.unit2internal("fs"),
        "eV": MDUnits.unit2internal("eV"),
        "atomic energy units": MDUnits.unit2internal("Ha"),
        "K": MDUnits.kB,
    }

    def __init__(self, start, stop=None, split=None):
        self.start = start
        self.stop = stop
        self.split = split
        self.read = False
        self.units = None
        self._matrix = []
        self._tmp_matrix = []

    def read_line(self, line):
        """
        Read and parse a line obtained from an open file object containing GLE parameters.

        Args:
            line (str): Line of a GLE parameter file.
        """
        line = line.strip()
        # Filter for empty lines
        if line:
            # Check if start token is present
            if self.start in line:
                self.read = True
                # Get units used
                unit_name = line.split("(")[-1].replace(")", "")
                self.units = self.unit_conversions[unit_name]
            elif self.read:
                if line.startswith("#"):
                    # Check for stop and split tokens
                    if self.stop is not None and self.stop in line:
                        self.read = False
                    if self.split is not None and self.split in line:
                        if len(self._tmp_matrix) > 0:
                            self._matrix.append(self._tmp_matrix)
                            self._tmp_matrix = []
                else:
                    # Otherwise read and parse line
                    self._tmp_matrix.append([float(x) for x in line.split()])

    @property
    def matrix(self):
        """
        Property to get parsed matrices converted to numpy arrays using atomic units.

        Returns:
            numpy.array: Array of the parsed GLE matrix with the shape normal_modes x s+1 x s+1, where normal_modes is 1
                         except in the case of the PIGLET thermostat and s is the number of degrees of freedom added via
                         GLE. If no matrix is found, None is returned.
        """
        # Write out last buffer
        if len(self._tmp_matrix) > 0:
            self._matrix.append(self._tmp_matrix)
        # Convert to numpy array
        _matrix = np.array(self._matrix)
        # Perform unit conversion
        if self.units is not None:
            return _matrix * self.units
        else:
            return None


class YSWeights:
    """
    Weights for Yoshida-Suzuki integration used in propagating the Nose-Hoover chain thermostats.

    Args:
        device (str): Device used for computation (default='cuda').
    """

    YS_weights = {
        3: np.array([1.35120719195966, -1.70241438391932, 1.35120719195966]),
        5: np.array(
            [
                0.41449077179438,
                0.41449077179438,
                -0.65796308717750,
                0.41449077179438,
                0.41449077179438,
            ]
        ),
        7: np.array(
            [
                0.78451361047756,
                0.23557321335936,
                -1.17767998417887,
                1.31518632068390,
                -1.17767998417887,
                0.23557321335936,
                0.78451361047756,
            ]
        ),
    }

    def __init__(self, device):
        self.device = device

    def get_weights(self, order):
        """
        Get the weights required for an integration scheme of the desired order.

        Args:
            order (int): Desired order of the integration scheme.

        Returns:
            torch.Tensor: Tensor of the integration weights
        """
        if order not in self.YS_weights:
            raise ValueError(
                "Order {:d} not supported for YS integration weights".format(order)
            )
        else:
            ys_weights = (
                torch.from_numpy(self.YS_weights[order]).float().to(self.device)
            )
        return ys_weights


class StableSinhDiv:
    """
    McLaurin series of sinh(x)/x around zero to avoid numerical instabilities
    """

    def __init__(self, eps=1e-4):
        self.e0 = 1.0
        self.e2 = self.e0 / 6.0
        self.e4 = self.e2 / 20.0
        self.e6 = self.e4 / 42.0
        self.e8 = self.e6 / 72.0
        self.eps = eps

    def f(self, x):
        if torch.min(torch.abs(x)) < self.eps:
            x = x ** 2
            return (((self.e8 * x + self.e6) * x + self.e4) * x + self.e2) * x + self.e0
        else:
            return torch.sinh(x) / x



import torch
import torch.nn as nn

from schnetpack.nn.base import Dense
from schnetpack import Properties
from schnetpack.nn.cfconv import CFConv
from schnetpack.nn.cutoff import CosineCutoff
from schnetpack.nn.acsf import GaussianSmearing
from schnetpack.nn.neighbors import AtomDistances
from schnetpack.nn.activations import shifted_softplus


__all__ = ["SchNetInteraction", "SchNet"]


class SchNetInteraction(nn.Module):
    r"""SchNet interaction block for modeling interactions of atomistic systems.

    Args:
        n_atom_basis (int): number of features to describe atomic environments.
        n_spatial_basis (int): number of input features of filter-generating networks.
        n_filters (int): number of filters used in continuous-filter convolution.
        cutoff (float): cutoff radius.
        cutoff_network (nn.Module, optional): cutoff layer.
        normalize_filter (bool, optional): if True, divide aggregated filter by number
            of neighbors over which convolution is applied.

    """

    def __init__(
        self,
        n_atom_basis,
        n_spatial_basis,
        n_filters,
        cutoff,
        cutoff_network=CosineCutoff,
        normalize_filter=False,
    ):
        super(SchNetInteraction, self).__init__()
        # filter block used in interaction block
        self.filter_network = nn.Sequential(
            Dense(n_spatial_basis, n_filters, activation=shifted_softplus),
            Dense(n_filters, n_filters),
        )
        # cutoff layer used in interaction block
        self.cutoff_network = cutoff_network(cutoff)
        # interaction block
        self.cfconv = CFConv(
            n_atom_basis,
            n_filters,
            n_atom_basis,
            self.filter_network,
            cutoff_network=self.cutoff_network,
            activation=shifted_softplus,
            normalize_filter=normalize_filter,
        )
        # dense layer
        self.dense = Dense(n_atom_basis, n_atom_basis, bias=True, activation=None)

    def forward(self, x, r_ij, neighbors, neighbor_mask, f_ij=None):
        """Compute interaction output.

        Args:
            x (torch.Tensor): input representation/embedding of atomic environments
                with (N_b, N_a, n_atom_basis) shape.
            r_ij (torch.Tensor): interatomic distances of (N_b, N_a, N_nbh) shape.
            neighbors (torch.Tensor): indices of neighbors of (N_b, N_a, N_nbh) shape.
            neighbor_mask (torch.Tensor): mask to filter out non-existing neighbors
                introduced via padding.
            f_ij (torch.Tensor, optional): expanded interatomic distances in a basis.
                If None, r_ij.unsqueeze(-1) is used.

        Returns:
            torch.Tensor: block output with (N_b, N_a, n_atom_basis) shape.

        """
        # continuous-filter convolution interaction block followed by Dense layer
        v = self.cfconv(x, r_ij, neighbors, neighbor_mask, f_ij)
        v = self.dense(v)
        return v


class SchNet(nn.Module):
    """SchNet architecture for learning representations of atomistic systems.

    Args:
        n_atom_basis (int, optional): number of features to describe atomic environments.
            This determines the size of each embedding vector; i.e. embeddings_dim.
        n_filters (int, optional): number of filters used in continuous-filter convolution
        n_interactions (int, optional): number of interaction blocks.
        cutoff (float, optional): cutoff radius.
        n_gaussians (int, optional): number of Gaussian functions used to expand
            atomic distances.
        normalize_filter (bool, optional): if True, divide aggregated filter by number
            of neighbors over which convolution is applied.
        coupled_interactions (bool, optional): if True, share the weights across
            interaction blocks and filter-generating networks.
        return_intermediate (bool, optional): if True, `forward` method also returns
            intermediate atomic representations after each interaction block is applied.
        max_z (int, optional): maximum nuclear charge allowed in database. This
            determines the size of the dictionary of embedding; i.e. num_embeddings.
        cutoff_network (nn.Module, optional): cutoff layer.
        trainable_gaussians (bool, optional): If True, widths and offset of Gaussian
            functions are adjusted during training process.
        distance_expansion (nn.Module, optional): layer for expanding interatomic
            distances in a basis.
        charged_systems (bool, optional):

    References:
    .. [#schnet1] Schütt, Arbabzadah, Chmiela, Müller, Tkatchenko:
       Quantum-chemical insights from deep tensor neural networks.
       Nature Communications, 8, 13890. 2017.
    .. [#schnet_transfer] Schütt, Kindermans, Sauceda, Chmiela, Tkatchenko, Müller:
       SchNet: A continuous-filter convolutional neural network for modeling quantum
       interactions.
       In Advances in Neural Information Processing Systems, pp. 992-1002. 2017.
    .. [#schnet3] Schütt, Sauceda, Kindermans, Tkatchenko, Müller:
       SchNet - a deep learning architecture for molceules and materials.
       The Journal of Chemical Physics 148 (24), 241722. 2018.

    """

    def __init__(
        self,
        n_atom_basis=128,
        n_filters=128,
        n_interactions=3,
        cutoff=5.0,
        n_gaussians=25,
        normalize_filter=False,
        coupled_interactions=False,
        return_intermediate=False,
        max_z=100,
        cutoff_network=CosineCutoff,
        trainable_gaussians=False,
        distance_expansion=None,
        charged_systems=False,
    ):
        super(SchNet, self).__init__()

        self.n_atom_basis = n_atom_basis
        # make a lookup table to store embeddings for each element (up to atomic
        # number max_z) each of which is a vector of size n_atom_basis
        self.embedding = nn.Embedding(max_z, n_atom_basis, padding_idx=0)

        # layer for computing interatomic distances
        self.distances = AtomDistances()

        # layer for expanding interatomic distances in a basis
        if distance_expansion is None:
            self.distance_expansion = GaussianSmearing(
                0.0, cutoff, n_gaussians, trainable=trainable_gaussians
            )
        else:
            self.distance_expansion = distance_expansion

        # block for computing interaction
        if coupled_interactions:
            # use the same SchNetInteraction instance (hence the same weights)
            self.interactions = nn.ModuleList(
                [
                    SchNetInteraction(
                        n_atom_basis=n_atom_basis,
                        n_spatial_basis=n_gaussians,
                        n_filters=n_filters,
                        cutoff_network=cutoff_network,
                        cutoff=cutoff,
                        normalize_filter=normalize_filter,
                    )
                ]
                * n_interactions
            )
        else:
            # use one SchNetInteraction instance for each interaction
            self.interactions = nn.ModuleList(
                [
                    SchNetInteraction(
                        n_atom_basis=n_atom_basis,
                        n_spatial_basis=n_gaussians,
                        n_filters=n_filters,
                        cutoff_network=cutoff_network,
                        cutoff=cutoff,
                        normalize_filter=normalize_filter,
                    )
                    for _ in range(n_interactions)
                ]
            )

        # set attributes
        self.return_intermediate = return_intermediate
        self.charged_systems = charged_systems
        if charged_systems:
            self.charge = nn.Parameter(torch.Tensor(1, n_atom_basis))
            self.charge.data.normal_(0, 1.0 / n_atom_basis ** 0.5)

    def forward(self, inputs):
        """Compute atomic representations/embeddings.

        Args:
            inputs (dict of torch.Tensor): SchNetPack dictionary of input tensors.

        Returns:
            torch.Tensor: atom-wise representation.
            list of torch.Tensor: intermediate atom-wise representations, if
            return_intermediate=True was used.

        """
        # get tensors from input dictionary
        atomic_numbers = inputs[Properties.Z]
        positions = inputs[Properties.R]
        cell = inputs[Properties.cell]
        cell_offset = inputs[Properties.cell_offset]
        neighbors = inputs[Properties.neighbors]
        neighbor_mask = inputs[Properties.neighbor_mask]
        atom_mask = inputs[Properties.atom_mask]

        # get atom embeddings for the input atomic numbers
        x = self.embedding(atomic_numbers)

        if self.charged_systems and Properties.charge in inputs.keys():
            n_atoms = torch.sum(atom_mask, dim=1, keepdim=True)
            charge = inputs[Properties.charge] / n_atoms  # B
            charge = charge[:, None] * self.charge  # B x F
            x = x + charge

        # compute interatomic distance of every atom to its neighbors
        r_ij = self.distances(
            positions, neighbors, cell, cell_offset, neighbor_mask=neighbor_mask
        )
        # expand interatomic distances (for example, Gaussian smearing)
        f_ij = self.distance_expansion(r_ij)
        # store intermediate representations
        if self.return_intermediate:
            xs = [x]
        # compute interaction block to update atomic embeddings
        for interaction in self.interactions:
            v = interaction(x, r_ij, neighbors, neighbor_mask, f_ij=f_ij)
            x = x + v
            if self.return_intermediate:
                xs.append(x)

        if self.return_intermediate:
            return x, xs
        return x



import os
import io

from setuptools import setup, find_packages


def read(fname):
    with io.open(os.path.join(os.path.dirname(__file__), fname), encoding="utf-8") as f:
        return f.read()


setup(
    name="schnetpack",
    version="1.0.1",
    author="Kristof T. Schuett, Michael Gastegger, Pan Kessel, Kim Nicoli",
    email="kristof.schuett@tu-berlin.de, michael.gastegger@tu-berlin.de",
    url="https://github.com/atomistic-machine-learning/schnetpack",
    packages=find_packages("src"),
    scripts=[
        "src/scripts/spk_ase.py",
        "src/scripts/spk_load.py",
        "src/scripts/spk_md.py",
        "src/scripts/spk_parse.py",
        "src/scripts/spk_run.py",
    ],
    package_dir={"": "src"},
    python_requires=">=3.6",
    install_requires=[
        "torch>=1.8.0",
        "numpy",
        "ase>=3.21",
        "h5py",
        "tensorboardX",
        "tqdm",
        "pyyaml",
    ],
    extras_require={"test": ["pytest", "pytest-console-scripts", "pytest-datadir"]},
    license="MIT",
    description="SchNetPack - Deep Neural Networks for Atomistic Systems",
    long_description="""
        SchNetPack aims to provide accessible atomistic neural networks that can be
        trained and applied out-of-the-box, while still being extensible to custom 
        atomistic architectures""",
)



from torch import nn as nn
import torch

from schnetpack import Properties

__all__ = ["AtomisticModel"]


class ModelError(Exception):
    pass


class AtomisticModel(nn.Module):
    """
    Join a representation model with output modules.

    Args:
        representation (torch.nn.Module): Representation block of the model.
        output_modules (list or nn.ModuleList or spk.output_modules.Atomwise): Output
            block of the model. Needed for predicting properties.

    Returns:
         dict: property predictions
    """

    def __init__(self, representation, output_modules):
        super(AtomisticModel, self).__init__()
        self.representation = representation
        if type(output_modules) not in [list, nn.ModuleList]:
            output_modules = [output_modules]
        if type(output_modules) == list:
            output_modules = nn.ModuleList(output_modules)
        self.output_modules = output_modules
        # For gradients
        self.requires_dr = any([om.derivative for om in self.output_modules])
        # For stress tensor
        self.requires_stress = any([om.stress for om in self.output_modules])

    def forward(self, inputs):
        """
        Forward representation output through output modules.
        """
        if self.requires_dr:
            inputs[Properties.R].requires_grad_()
        if self.requires_stress:
            # Check if cell is present
            if inputs[Properties.cell] is None:
                raise ModelError("No cell found for stress computation.")

            # Generate Cartesian displacement tensor
            displacement = torch.zeros_like(inputs[Properties.cell]).to(
                inputs[Properties.R].device
            )
            displacement.requires_grad = True
            inputs["displacement"] = displacement

            # Apply to coordinates and cell
            inputs[Properties.R] = inputs[Properties.R] + torch.matmul(
                inputs[Properties.R], displacement
            )
            inputs[Properties.cell] = inputs[Properties.cell] + torch.matmul(
                inputs[Properties.cell], displacement
            )

        inputs["representation"] = self.representation(inputs)
        outs = {}
        for output_model in self.output_modules:
            outs.update(output_model(inputs))
        return outs



import torch
import os
import subprocess
from ase import Atoms
from ase.data import chemical_symbols

from schnetpack.md.calculators.basic_calculators import (
    QMCalculatorError,
    QMCalculator,
    Queuer,
)
from schnetpack.md.utils import MDUnits
from schnetpack.md.parsers.orca_parser import OrcaMainFileParser
from schnetpack import Properties


class OrcaCalculator(QMCalculator):
    """
    Calculator for interfacing the ORCA code package [#orca1]_ with SchNetPack molecular dynamics.
    Requires ORCA to be installed and an input file template.
    This template is a standard ORCA input file, with everything past the specification of coordinate
    format, charge and multiplicity removed (coordinates and final `*`).
    If desired, a Queuer can be give, which will attempt to send all jobs to a grid engine queue.

    In general, the calculator will take the current System to generate inputs, perform the calculation
    with ORCA, extract data from the ouput file (useing the OrcaParser class) and update the System.

    Args:
        required_properties (list): List of properties which should be extracted from output.
        force_handle (str): Indicator for molecular forces.
        compdir (str): Directory in which computations are performed.
        qm_executable (str): Path to the ORCA executable.
        orca_template (str): Path to an ORCA template which will be used to generate input files.
        orca_parser (schnetpack.md.parsers.OrcaParser, optional): Parser used to extract data from output files.
        position_conversion (str/float, optional): Conversion of positions from atomic units.
        force_conversion (str/float, optional): Conversion of forces to atomic units.
        property_conversion (dict, optional): Convert properties to requested units. If left empty, no conversion
                                              is performed.
        queuer (schnetpack.md.calculator.Queuer, optional): If given, jobs will be submitted to a grid engine queue.
        adaptive (bool, optional): Specify, whether the calculator should be used for adaptive sampling.
        basename (str, optional): Basename of the generated input files.

    References
    ----------
    .. [#orca1] Neese:
       The ORCA program system.
       WIREs Comput Mol Sci, 2 (1), 73-78. 2012.
    """

    is_atomistic = [Properties.forces, Properties.shielding]

    def __init__(
        self,
        required_properties,
        force_handle,
        compdir,
        qm_executable,
        orca_template,
        orca_parser=OrcaMainFileParser,
        position_conversion="Angstrom",
        force_conversion="Ha/Bohr",
        property_conversion={},
        queuer=None,
        adaptive=False,
        basename="input",
    ):
        super(OrcaCalculator, self).__init__(
            required_properties,
            force_handle,
            compdir,
            qm_executable,
            position_conversion=position_conversion,
            force_conversion=force_conversion,
            property_conversion=property_conversion,
            adaptive=adaptive,
        )

        self.orca_template = open(orca_template, "r").read()
        self.basename = basename
        self.orca_parser = orca_parser(properties=required_properties)

        self.queuer = queuer

    def _generate_orca_inputs(self, molecules, current_compdir):
        """
        Generate input files for all molecules in the current System.

        Args:
            molecules (list of ase.Atoms): List of ASE Atoms objects containing the structures and atom types
                                           of all molecules to be computed.
            current_compdir (str): Path to the computation directory in which the input files should be generated.

        Returns:
            list: List of all generated input files.
        """
        input_files = []
        for idx, molecule in enumerate(molecules):
            # Convert data and generate input files
            atom_types, positions = molecule
            input_file_name = os.path.join(
                current_compdir, "{:s}_{:06d}.oinp".format(self.basename, idx + 1)
            )
            self._write_orca_input(input_file_name, atom_types, positions)
            input_files.append(input_file_name)

        return input_files

    def _write_orca_input(self, input_file_name, atom_types, positions):
        """
        Write the ORCA input file using the provided template.

        Args:
            input_file_name (str): Name of the input file.
            atom_types (numpy.array): Integer array holding the atomic numbers of each species.
            positions (numpy.array): Array of all atom positions in Angstrom.
        """
        input_file = open(input_file_name, "w")
        input_file.write(self.orca_template)
        for idx in range(len(atom_types)):
            input_file.write(
                "{:2s} {:15.8f} {:15.8f} {:15.8f}\n".format(
                    chemical_symbols[atom_types[idx]], *positions[idx]
                )
            )
        input_file.write("*")
        input_file.close()

    def _run_computation(self, molecules, current_compdir):
        """
        Perform the actual computation.
        First, inputs are generated, then computations are performed, finally the data is extracted
        from the output files and returned.

        Args:
            molecules (list): List of atom_types, position tuples
            current_compdir (str): Path to computation directory.

        Returns:
            dict: Dictionary holding the requested properties which were computed.
        """
        # Generate input files
        input_files = self._generate_orca_inputs(molecules, current_compdir)

        # Perform computations
        if self.queuer is None:
            for input_file in input_files:
                command = "{:s} {:s}".format(self.qm_executable, input_file, input_file)
                with open("{:s}.log".format(input_file), "wb") as out:
                    computation = subprocess.Popen(command.split(), stdout=out)
                    computation.wait()
        else:
            self.queuer.submit(input_files, current_compdir)

        # Extract the results
        outputs = []
        for input_file in input_files:
            self.orca_parser.parse_file("{:s}.log".format(input_file))

            orca_outputs = self.orca_parser.get_parsed()

            for p in self.required_properties:
                if orca_outputs[p] is None:
                    raise QMCalculatorError(
                        "Requested property {:s} was not computed in {:s}".format(
                            p, input_file
                        )
                    )

            outputs.append(orca_outputs)

        return outputs

    def _format_calc(self, outputs, system):
        """
        Format the extracted properties into the form used by the schnetpack.md.System
        class (zero padding, reshaping, etc.).

        Args:
            outputs (list): Paths to output files.
            system (schnetpack.md.System): System from the molecular dynamics simulation.

        Returns:
            dict: Dictionary of all extracted computation results formatted and padded to the
                  right dimensions used in the system class.
        """
        max_natoms = system.max_n_atoms
        results = {p: [] for p in self.required_properties}

        for output in outputs:
            for p in self.required_properties:
                # Check for convergence
                if output[p] is None:
                    raise QMCalculatorError("Errors encountered during computation.")
                if p in self.is_atomistic:
                    padded_output = torch.zeros(max_natoms, *output[p].shape[1:])
                    padded_output[: output[p].shape[0], ...] = torch.from_numpy(
                        output[p]
                    )
                    results[p].append(padded_output)
                else:
                    results[p].append(torch.from_numpy(output[p]))

        for p in self.required_properties:
            results[p] = torch.stack(results[p]).to(system.device)

        return results

    def _format_ase(self, molecules, outputs):
        """
        Format the ouput for storage in an ASE database (for adaptive sampling).

        Args:
            molecules (list): List of tuples containing the atom types (integer numpy.array)
                      and positions (float numpy.array).
            outputs (list): Paths to output files.
        """
        atom_buffer = []
        property_buffer = []
        for idx, molecule in enumerate(molecules):
            atom_types, positions = molecule
            atoms = Atoms(atom_types.cpu(), positions.cpu())

            properties = outputs[idx]

            atom_buffer.append(atoms)
            property_buffer.append(properties)
        return atom_buffer, property_buffer


class OrcaQueuer(Queuer):
    """
    Class for interfacing the ORCA calculator with a grid engine queue.
    Will create the directory structure and a submission command, which it will pass to the
    given queue. Jobs will be submitted as array jobs, where the number of concurrently running
    jobs can be changed. Should be adapted to work with the individual grid setups.

    Args:
        queue (str): Indentifier for the queue the job should be sent to.
        orca_executable (str): Path to the executable to be called.
        concurrent (str): How many concurrent jobs should be run in each array. (default=100)
        basename (str): Basic identifier used for job (default='input')
        cleanup (bool): Whether directories should be deleted (default=True)
    """

    QUEUE_FILE = """
#!/usr/bin/env bash
##############################
#$ -cwd
#$ -V
#$ -q {queue}
#$ -N {jobname}
#$ -t 1-{array_range}
#$ -tc {concurrent}
#$ -S /bin/bash
#$ -e /dev/null
#$ -o /dev/null
#$ -r n
#$ -sync y
##############################

task_name={basename}_$(printf "%06d" $SGE_TASK_ID)

export QSUB_WORKDIR=/tmp/$USER/$task_name.$$
mkdir -p $QSUB_WORKDIR

cp {compdir}/$task_name.oinp $QSUB_WORKDIR
cd $QSUB_WORKDIR

{orca_path} $task_name.oinp > $task_name.oinp.log

cp -f $QSUB_WORKDIR/$task_name* {compdir}/
rm -rf $QSUB_WORKDIR
"""

    def __init__(
        self, queue, orca_executable, concurrent=100, basename="input", cleanup=True
    ):
        super(OrcaQueuer, self).__init__(
            queue,
            orca_executable,
            concurrent=concurrent,
            basename=basename,
            cleanup=cleanup,
        )

    def _create_submission_command(self, n_inputs, compdir, jobname):
        """
        Use the QUEUE_FILE template to creates the submission command.

        Args:
            n_inputs (int): Number of input files to be submitted
            compdir (str): Current working directory
            jobname (str): Name of each job

        Returns:
            str: Submission command.
        """
        submission_command = self.QUEUE_FILE.format(
            queue=self.queue,
            basename=self.basename,
            array_range=n_inputs,
            concurrent=self.concurrent,
            compdir=compdir,
            orca_path=self.executable,
            jobname=jobname,
        )
        return submission_command



import torch
from torch import nn
from torch.nn.init import xavier_uniform_

from schnetpack.nn.initializers import zeros_initializer


__all__ = ["Dense", "GetItem", "ScaleShift", "Standardize", "Aggregate"]


class Dense(nn.Linear):
    r"""Fully connected linear layer with activation function.

    .. math::
       y = activation(xW^T + b)

    Args:
        in_features (int): number of input feature :math:`x`.
        out_features (int): number of output features :math:`y`.
        bias (bool, optional): if False, the layer will not adapt bias :math:`b`.
        activation (callable, optional): if None, no activation function is used.
        weight_init (callable, optional): weight initializer from current weight.
        bias_init (callable, optional): bias initializer from current bias.

    """

    def __init__(
        self,
        in_features,
        out_features,
        bias=True,
        activation=None,
        weight_init=xavier_uniform_,
        bias_init=zeros_initializer,
    ):
        self.weight_init = weight_init
        self.bias_init = bias_init
        self.activation = activation
        # initialize linear layer y = xW^T + b
        super(Dense, self).__init__(in_features, out_features, bias)

    def reset_parameters(self):
        """Reinitialize model weight and bias values."""
        self.weight_init(self.weight)
        if self.bias is not None:
            self.bias_init(self.bias)

    def forward(self, inputs):
        """Compute layer output.

        Args:
            inputs (dict of torch.Tensor): batch of input values.

        Returns:
            torch.Tensor: layer output.

        """
        # compute linear layer y = xW^T + b
        y = super(Dense, self).forward(inputs)
        # add activation function
        if self.activation:
            y = self.activation(y)
        return y


class GetItem(nn.Module):
    """Extraction layer to get an item from SchNetPack dictionary of input tensors.

    Args:
        key (str): Property to be extracted from SchNetPack input tensors.

    """

    def __init__(self, key):
        super(GetItem, self).__init__()
        self.key = key

    def forward(self, inputs):
        """Compute layer output.

        Args:
            inputs (dict of torch.Tensor): SchNetPack dictionary of input tensors.

        Returns:
            torch.Tensor: layer output.

        """
        return inputs[self.key]


class ScaleShift(nn.Module):
    r"""Scale and shift layer for standardization.

    .. math::
       y = x \times \sigma + \mu

    Args:
        mean (torch.Tensor): mean value :math:`\mu`.
        stddev (torch.Tensor): standard deviation value :math:`\sigma`.

    """

    def __init__(self, mean, stddev):
        super(ScaleShift, self).__init__()
        self.register_buffer("mean", mean)
        self.register_buffer("stddev", stddev)

    def forward(self, input):
        """Compute layer output.

        Args:
            input (torch.Tensor): input data.

        Returns:
            torch.Tensor: layer output.

        """
        y = input * self.stddev + self.mean
        return y


class Standardize(nn.Module):
    r"""Standardize layer for shifting and scaling.

    .. math::
       y = \frac{x - \mu}{\sigma}

    Args:
        mean (torch.Tensor): mean value :math:`\mu`.
        stddev (torch.Tensor): standard deviation value :math:`\sigma`.
        eps (float, optional): small offset value to avoid zero division.

    """

    def __init__(self, mean, stddev, eps=1e-9):
        super(Standardize, self).__init__()
        self.register_buffer("mean", mean)
        self.register_buffer("stddev", stddev)
        self.register_buffer("eps", torch.ones_like(stddev) * eps)

    def forward(self, input):
        """Compute layer output.

        Args:
            input (torch.Tensor): input data.

        Returns:
            torch.Tensor: layer output.

        """
        # Add small number to catch divide by zero
        y = (input - self.mean) / (self.stddev + self.eps)
        return y


class Aggregate(nn.Module):
    """Pooling layer based on sum or average with optional masking.

    Args:
        axis (int): axis along which pooling is done.
        mean (bool, optional): if True, use average instead for sum pooling.
        keepdim (bool, optional): whether the output tensor has dim retained or not.

    """

    def __init__(self, axis, mean=False, keepdim=True):
        super(Aggregate, self).__init__()
        self.average = mean
        self.axis = axis
        self.keepdim = keepdim

    def forward(self, input, mask=None):
        r"""Compute layer output.

        Args:
            input (torch.Tensor): input data.
            mask (torch.Tensor, optional): mask to be applied; e.g. neighbors mask.

        Returns:
            torch.Tensor: layer output.

        """
        # mask input
        if mask is not None:
            input = input * mask[..., None]
        # compute sum of input along axis
        y = torch.sum(input, self.axis)
        # compute average of input along axis
        if self.average:
            # get the number of items along axis
            if mask is not None:
                N = torch.sum(mask, self.axis, keepdim=self.keepdim)
                N = torch.max(N, other=torch.ones_like(N))
            else:
                N = input.size(self.axis)
            y = y / N
        return y


class MaxAggregate(nn.Module):
    """Pooling layer that computes the maximum for each feature over all atoms

    Args:
        axis (int): axis along which pooling is done.
    """

    def __init__(self, axis):
        super().__init__()
        self.axis = axis

    def forward(self, input, mask=None):
        r"""Compute layer output.

        Args:
            input (torch.Tensor): input data.
            mask (torch.Tensor, optional): mask to be applied; e.g. neighbors mask.

        Returns:
            torch.Tensor: layer output.
        """
        # mask input
        if mask is not None:
            # If the mask is lower dimensional than the array being masked,
            #  inject an extra dimension to the end
            if mask.dim() < input.dim():
                mask = torch.unsqueeze(mask, -1)
            input = torch.where(mask > 0, input, torch.min(input))

        # compute sum of input along axis
        return torch.max(input, self.axis)[0]


class SoftmaxAggregate(nn.Module):
    """Pooling layer that computes the maximum for each feature over all atoms
    using the "softmax" function to weigh the contribution of each atom to
    the "maximum."

    Args:
        axis (int): axis along which pooling is done.
    """

    def __init__(self, axis):
        super().__init__()
        self.axis = axis

    def forward(self, input, mask=None):
        r"""Compute layer output.

        Args:
            input (torch.Tensor): input data.
            mask (torch.Tensor, optional): mask to be applied; e.g. neighbors mask.

        Returns:
            torch.Tensor: layer output.
        """

        # Compute the sum of exponentials for the desired axis
        exp_input = torch.exp(input)

        # Set the contributions of "masked" atoms to zero
        if mask is not None:
            # If the mask is lower dimensional than the array being masked,
            #  inject an extra dimension to the end
            if mask.dim() < input.dim():
                mask = torch.unsqueeze(mask, -1)
            exp_input = torch.where(mask > 0, exp_input, torch.zeros_like(exp_input))

        # Sum exponentials along the desired axis
        exp_input_sum = torch.sum(exp_input, self.axis, keepdim=True)

        # Normalize the exponential array by the
        weights = exp_input / exp_input_sum

        # compute sum of input along axis
        output = torch.sum(input * weights, self.axis)
        return output



class Hook:
    """Base class for hooks."""

    @property
    def state_dict(self):
        return {}

    @state_dict.setter
    def state_dict(self, state_dict):
        pass

    def on_train_begin(self, trainer):
        pass

    def on_train_ends(self, trainer):
        pass

    def on_train_failed(self, trainer):
        pass

    def on_epoch_begin(self, trainer):
        """Log at the beginning of train epoch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.

        """
        pass

    def on_batch_begin(self, trainer, train_batch):
        """Log at the beginning of train batch.

        Args:
            trainer (Trainer): instance of schnetpack.train.trainer.Trainer class.
            train_batch (dict of torch.Tensor): SchNetPack dictionary of input tensors.

        """
        pass

    def on_batch_end(self, trainer, train_batch, result, loss):
        pass

    def on_validation_begin(self, trainer):
        pass

    def on_validation_batch_begin(self, trainer):
        pass

    def on_validation_batch_end(self, trainer, val_batch, val_result):
        pass

    def on_validation_end(self, trainer, val_loss):
        pass

    def on_epoch_end(self, trainer):
        pass



"""
Collection of utilities for computing autocorrelation functions and molecular spectra
from HDF5 files generated during molecular dynamics. For a good overview on how to compute
spectra from molecular dynamics simulations and details on the techniques used, we recommend [#spectra1]_ .

References
----------
.. [#spectra1] Martin, Brehm, Fligg, Vöhringer, Kirchner:
               Computing vibrational spectra from ab initio molecular dynamics.
               Phys. Chem. Chem. Phys., 15 (18), 6608--6622. 2013.
"""
import numpy as np
from ase import units
from schnetpack.md.utils import MDUnits
from schnetpack import Properties
import logging


def cosine_sq_window(n_points):
    """
    Squared cosine window function for spectra.

    Args:
        n_points (int): Number of points in spectrum

    Returns:
        numpy.array: Squared cosine window function.
    """
    points = np.arange(n_points)
    window = np.cos(np.pi * points / (n_points - 1) / 2) ** 2
    return window


def fft_autocorrelation(data, n_lags):
    """
    Routine for fast computation of autocorrelation using FFT and Wiener--Kinchie theorem.

    Args:
        data (numpy.array): Array containing data for which autocorrelation should be computed.
        n_lags (int): Number of time lags used for extracting the autocorrelation.

    Returns:
        numpy.array: Autocorrelation function of the input array
    """
    # Normalize data to get autocorrelation[0] = 1
    data = (data - np.mean(data)) / np.std(data)
    n_points = data.shape[0]
    fft_forward = np.fft.fft(data, n=2 * n_points)
    fft_autocorr = fft_forward * np.conjugate(fft_forward)
    fft_backward = np.fft.ifft(fft_autocorr)[:n_points] / n_points
    autocorrelation = np.real(fft_backward[: n_lags + 1])
    return autocorrelation


class VibrationalSpectrum:
    """
    Base class for computing vibrational spectra from HDF5 datasets using autocorrelation functions and
    fast fourier transforms.

    Args:
        data (schnetpack.md.utils.HDF5Loader): Loaded dataset.
        resolution (int): Resolution used when computing the spectrum. Indicates how many time lags are considered
                          in the autocorrelation function is used.
        window (function, optional): Window function used for computing the spectrum.
    """

    def __init__(self, data, resolution=4096, window=cosine_sq_window):
        self.data = data
        self.timestep = data.time_step / MDUnits.fs2internal  # Convert to fs
        self.resolution = resolution
        self.window = window

        spectral_range = 0.5 / self.timestep / (units._c / 1e13)
        spectral_resolution = spectral_range / (4 * resolution)
        logging.info(
            "Spectral resolutions: {:12.3f} [cm^-1]".format(spectral_resolution)
        )
        logging.info("Spectral range:       {:12.3f} [cm^-1]".format(spectral_range))

        self.res = spectral_resolution
        self.frequencies = []
        self.intensities = []

    def compute_spectrum(self, molecule_idx=0):
        """
        Main routine for computing spectra. First the relavant data is read,
        then autocorrelations are computed and processed. Based on the
        processed autocorrelations, spectra are computed and, if requested,
        subjected to additional postprocessing.

        Args:
            molecule_idx (int): Index of the molecule for which the spectrum should be computed.
                                Uses the same conventions as schnetpack.md.System.
        """
        # Get appropriate data
        relevant_data = self._get_data(molecule_idx)

        # Compute autocorrelation function
        autocorrelation = self._compute_autocorrelations(relevant_data)
        # Process the autocorrelation function (e.g. weighting, Raman, ...)
        autocorrelation = self._process_autocorrelation(autocorrelation)

        self.frequencies = []
        self.intensities = []
        # Compute spectrum
        for autocorr in autocorrelation:
            frequencies, intensities = self._compute_spectrum(autocorr)
            self.frequencies.append(frequencies)
            self.intensities.append(intensities)

        self._process_spectrum()

    def _compute_spectrum(self, autocorrelation):
        """
        Compute the spectrum from the autocorrelation function.

        Args:
            autocorrelation (numpy.array): Autorcorrelation function.

        Returns:
            (numpy.array,numpy.array):
                frequencies:
                    Vibrational frequencies in inverse centimeters.
                intensities:
                    Intensities of the vibrational bands.
        """
        data = autocorrelation[: self.resolution]

        # Various tricks for nicer spectra
        # 1) Apply window function
        n_unpadded = data.shape[0]
        if self.window is not None:
            data *= self.window(n_unpadded)
        # 2) Zero padding
        data_padded = np.zeros(4 * n_unpadded)
        data_padded[:n_unpadded] = data
        # 3) Mirror around origin
        data_mirrored = np.hstack((np.flipud(data_padded), data_padded))
        # Compute the spectrum
        n_fourier = 8 * n_unpadded
        intensities = np.abs(
            self.timestep * np.fft.fft(data_mirrored, n=n_fourier)[: n_fourier // 2]
        )
        frequencies = np.arange(n_fourier // 2) / (n_fourier * self.timestep)
        # Conversion to inverse cm
        frequencies /= units._c / 1e13
        return frequencies, intensities

    @staticmethod
    def _compute_autocorrelations(data):
        """
        Compute the autocorrelation function of the data. A separate autocorrelation is computred
        for every array dimension except the first axis.

        Args:
            data (numpy.array): Function array.

        Returns:
            numpy.array: Autocorrelation of the inputs.
        """
        n_data = data.shape[0]
        data_dim = data.shape[1:]
        n_lags = n_data - 2

        # Flatten data for easier iteration
        reshaped_data = data.reshape((n_data, -1))
        n_fields = reshaped_data.shape[1]

        # Compute all individual autocorrelations
        autocorrelations = np.zeros((n_fields, n_lags + 1))
        for field in range(n_fields):
            autocorrelations[field, ...] = fft_autocorrelation(
                reshaped_data[..., field], n_lags
            )

        # Reconstruct shape of original property
        autocorrelations = autocorrelations.reshape((*data_dim, -1))
        return autocorrelations

    def _get_data(self, molecule_idx):
        """
        Placeholder for extracting teh required data from the HDF5 dataset.

        Args:
            molecule_idx (int): Index of the molecule for which the spectrum should be computed.
                                Uses the same conventions as schnetpack.md.System.
        """
        raise NotImplementedError

    def _process_autocorrelation(self, autocorrelation):
        """
        Placeholder for postprocessing the autocorrelation functions (e.g. weighting).

        Args:
            autocorrelation (numpy.array): Autorcorrelation function.
        """
        raise NotImplementedError

    def _process_spectrum(self):
        """
        Placeholder function if postprocessing should be applied to the spectrum (e.g. quantum coorections).
        """
        pass

    def get_spectrum(self):
        """
        Returns all computed spectra in the form of a list of tuples of frequencies and intensities.

        Returns:
            list: List of tuples of frequencies and intensities of all computed spectra.
        """
        spectrum = list(zip(self.frequencies, self.intensities))
        if len(spectrum) == 1:
            return spectrum[0]
        else:
            return spectrum


class PowerSpectrum(VibrationalSpectrum):
    """
    Compute power spectra from a molecular dynamics HDF5 dataset.

    Args:
        data (schnetpack.md.utils.HDF5Loader): Loaded dataset.
        resolution (int, optional): Resolution used when computing the spectrum. Indicates how many time lags
                                    are considered in the autocorrelation function is used.
    """

    def __init__(self, data, resolution=4096):
        super(PowerSpectrum, self).__init__(data, resolution=resolution)

    def _get_data(self, molecule_idx):
        """
        Extract the molecular velocities for computing the spectrum.

        Args:
            molecule_idx (int): Index of the molecule for which the spectrum should be computed.
                                Uses the same conventions as schnetpack.md.System.

        Returns:
            numpy.array: Array holding molecular velocities.
        """
        relevant_data = self.data.get_velocities(molecule_idx)
        return relevant_data

    def _process_autocorrelation(self, autocorrelation):
        """
        Sum over number of atoms and the three Cartesian components.

        Args:
            autocorrelation (numpy.array): Autorcorrelation function.

        Returns:
            numpy.array: Updated autocorrelation.
        """
        vdos_autocorrelation = np.sum(autocorrelation, axis=1)
        vdos_autocorrelation = np.mean(vdos_autocorrelation, axis=0)
        return [vdos_autocorrelation]


class IRSpectrum(VibrationalSpectrum):
    """
    Compute infrared spectra from a molecular dynamics HDF5 dataset. This class requires the dipole moments
    to be present in the HDF5 dataset.

    Args:
        data (schnetpack.md.utils.HDF5Loader): Loaded dataset.
        resolution (int, optional): Resolution used when computing the spectrum. Indicates how many time lags
                                    are considered in the autocorrelation function is used.
        dipole_moment_handle (str, optional): Indentifier used for extracting dipole data.
    """

    def __init__(
        self, data, resolution=4096, dipole_moment_handle=Properties.dipole_moment
    ):
        super(IRSpectrum, self).__init__(data, resolution=resolution)
        self.dipole_moment_handle = dipole_moment_handle

    def _get_data(self, molecule_idx):
        """
        Extract the molecular dipole moments and compute their time derivative via central
        difference.

        Args:
            molecule_idx (int): Index of the molecule for which the spectrum should be computed.
                                Uses the same conventions as schnetpack.md.System.

        Returns:
            numpy.array: Array holding the dipole moment derivatives.
        """
        relevant_data = self.data.get_property(self.dipole_moment_handle, molecule_idx)
        # Compute numerical derivative via central differences
        relevant_data = (relevant_data[2:, ...] - relevant_data[:-2, ...]) / (
            2 * self.timestep
        )
        return relevant_data

    def _process_autocorrelation(self, autocorrelation):
        """
        Sum over the three Cartesian components.

        Args:
            autocorrelation (numpy.array): Dipole moment flux autorcorrelation functions.

        Returns:
            numpy.array: Updated autocorrelation.
        """
        dipole_autocorrelation = np.sum(autocorrelation, axis=0)
        return [dipole_autocorrelation]



import logging
import os
import shutil
import tempfile
from urllib import request as request

import numpy as np
from ase import Atoms

import schnetpack as spk
from schnetpack.data import AtomsDataError
from schnetpack.datasets import DownloadableAtomsData

__all__ = ["MD17"]


class MD17(DownloadableAtomsData):
    """
    MD17 benchmark data set for molecular dynamics of small molecules
    containing molecular forces.

    Args:
        dbpath (str): path to database
        molecule (str or None): Name of molecule to load into database. Allowed are:
                            aspirin
                            benzene
                            ethanol
                            malonaldehyde
                            naphthalene
                            salicylic_acid
                            toluene
                            uracil
        subset (list, optional): Deprecated! Do not use! Subsets are created with
            AtomsDataSubset class.
        download (bool): set true if dataset should be downloaded
            (default: True)
        collect_triples (bool): set true if triples for angular functions
            should be computed (default: False)
        load_only (list, optional): reduced set of properties to be loaded
        environment_provider (spk.environment.BaseEnvironmentProvider): define how
            neighborhood is calculated
            (default=spk.environment.SimpleEnvironmentProvider).


    See: http://quantum-machine.org/datasets/
    """

    energy = "energy"
    forces = "forces"

    datasets_dict = dict(
        aspirin="aspirin_dft.npz",
        # aspirin_ccsd='aspirin_ccsd.zip',
        azobenzene="azobenzene_dft.npz",
        benzene="benzene_dft.npz",
        ethanol="ethanol_dft.npz",
        # ethanol_ccsdt='ethanol_ccsd_t.zip',
        malonaldehyde="malonaldehyde_dft.npz",
        # malonaldehyde_ccsdt='malonaldehyde_ccsd_t.zip',
        naphthalene="naphthalene_dft.npz",
        paracetamol="paracetamol_dft.npz",
        salicylic_acid="salicylic_dft.npz",
        toluene="toluene_dft.npz",
        # toluene_ccsdt='toluene_ccsd_t.zip',
        uracil="uracil_dft.npz",
    )

    existing_datasets = datasets_dict.keys()

    def __init__(
        self,
        dbpath,
        molecule=None,
        subset=None,
        download=True,
        collect_triples=False,
        load_only=None,
        environment_provider=spk.environment.SimpleEnvironmentProvider(),
    ):
        if not os.path.exists(dbpath) and molecule is None:
            raise AtomsDataError("Provide a valid dbpath or select desired molecule!")

        if molecule is not None and molecule not in MD17.datasets_dict.keys():
            raise AtomsDataError("Molecule {} is not supported!".format(molecule))

        self.molecule = molecule

        available_properties = [MD17.energy, MD17.forces]

        super(MD17, self).__init__(
            dbpath=dbpath,
            subset=subset,
            load_only=load_only,
            collect_triples=collect_triples,
            download=download,
            available_properties=available_properties,
            environment_provider=environment_provider,
        )

    def _download(self):

        logging.info("Downloading {} data".format(self.molecule))
        tmpdir = tempfile.mkdtemp("MD")
        rawpath = os.path.join(tmpdir, self.datasets_dict[self.molecule])
        url = (
            "http://www.quantum-machine.org/gdml/data/npz/"
            + self.datasets_dict[self.molecule]
        )

        request.urlretrieve(url, rawpath)

        logging.info("Parsing molecule {:s}".format(self.molecule))

        data = np.load(rawpath)

        numbers = data["z"]
        atoms_list = []
        properties_list = []
        for positions, energies, forces in zip(data["R"], data["E"], data["F"]):
            properties_list.append(dict(energy=energies, forces=forces))
            atoms_list.append(Atoms(positions=positions, numbers=numbers))

        self.add_systems(atoms_list, properties_list)
        self.update_metadata(dict(data_source=self.datasets_dict[self.molecule]))

        logging.info("Cleanining up the mess...")
        logging.info("{} molecule done".format(self.molecule))
        shutil.rmtree(tmpdir)
